from sqlalchemy import create_engine
from common import *
db = create_engine('postgresql://localhost/dbwipes')
sigmoddb = create_engine('postgresql://localhost/sigmod')

def wrap(f):
    def _f(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except:
            import traceback
            traceback.print_exc()
            import pdb
            pdb.set_trace()
    return _f
            



def get_ids_from_bounds(dataset, bounds):
    """ bounds: [ (dimsuffix, lower, upper)* ] """

    wheres = []
    for d,l,u in bounds:
        if not math.isinf(l):
            wheres.append("%f <= a%d" % (l, d))
        if not math.isinf(u):
            wheres.append("a%d <= %f" % (d, u))
    where = ' and '.join(wheres) if wheres else ' 1 = 1 '
    with sigmoddb.begin() as sconn:
        q = "select id from %s where %s" % (dataset, where)
        ret = []
        for (id,) in sconn.execute(q):
            ret.append(int(id))

    return set(ret)

@wrap
def get_truths():

    db = create_engine('postgresql://localhost/dbwipes')
    q = """
    select distinct dataset, boundtype, notes from stats  where substring(dataset from 1 for 1) = 'd'   """
#    q = """
#    select s1.dataset, s1.expid, s1.c, s1.cost, s1.ids
#    from stats as s1, (select expid, c, max(cost) as cost
#                       from stats as s2
#                       where s2.klass = 'Naive'
#                       group by expid, c) as maxes
#    where klass = 'Naive' and 
#          s1.expid = maxes.expid and
#          s1.c = maxes.c and 
#          s1.cost = maxes.cost
#    """

    with db.begin() as conn:
        for dataset, boundtype, notes in conn.execute(q):
            pts = map(float,notes.replace(' ', '').replace('[','').replace(']', '').split(','))
            bounds = [(dim, pts[dim*2], pts[dim*2+1]) for dim in xrange(len(pts) / 2)]

            yield dataset, boundtype, get_ids_from_bounds(dataset, bounds)
            

cache = {}

@wrap
def get_alternative_results(db, did, boundtype):
    q = """
    select klass, id, c, cost, rule
    from stats
    where dataset = %s and boundtype = %s and klass = 'MR'
    order by klass, c, cost asc
    """
    with db.begin() as conn:
        for klass, id, c, cost, rule in conn.execute(q, did, boundtype).fetchall():
            key = hash((klass, c, did, rule))
            if key in cache:
                yield id, c, cost, cache[key]

            if rule:
                dimstrs = rule.split(' and ')
                dimstrs = [ds.split(' ') for ds in dimstrs]
                bounds = [(int(arr[2][-1]), float(arr[0]), float(arr[-1])) for arr in dimstrs]
            else:
                bounds = []

            cache[key] = get_ids_from_bounds(did, bounds)

            yield id, c, cost, cache[key] 

def save_scores(db, rowid, acc, p, r, f1):
    q = """
    update stats set acc = %s, prec = %s, recall = %s, f1 = %s
    where id = %s
    """
    with db.begin() as conn:
        conn.execute(q, acc, p, r, f1, rowid)

def table_size(dbname, table):
    db = create_engine('postgresql://localhost/%s' % dbname)
    with db.begin() as conn:
        return conn.execute("select count(*) from %s" % table).fetchone()[0]

def dataset_to_info(did):
    if did == 0:
        return ('intel', 'readings')
    if did == 5:
        return ('harddata', 'harddata_1')
    if did == 11:
        return ('fec12', 'expenses')
    if did == 15:
        return ('harddata','multdim') 
    return ('sigmod', did)

for dataset, boundtype, true_ids in get_truths():
    dbname, tname = dataset_to_info(dataset)
    print dbname, tname, boundtype
    tsize = table_size(dbname, tname)
    #for rowid, cost, ids in  get_checkpoints(db, eid, c, max_cost):
    for rowid, c, cost, ids in get_alternative_results(db, dataset, boundtype):
        acc, p, r, f1 = compute_stats(ids, true_ids, tsize)
        #print '\t%.4f\t%.4f\t%.4f\t%.4f' % ( cost, p, r, f1)
        save_scores(db, rowid, acc, p, r, f1)






import sys
import numpy as np
import Orange
import orange
import math
import random
import pdb
sys.path.extend(['.', '..'])

from collections import Counter, defaultdict
from itertools import chain
from scorpionsql.aggerror import *

from ..settings import *
from ..util import *
from ..learners.cn2sd.refiner import *

_logger = get_logger()
inf = float('inf')


class Sampler(object):
  def __init__(self, SCORE_ID):
    self.SCORE_ID = SCORE_ID

  def get_sample_size(self, pop, *args, **kwargs):
    return best_sample_size(pop, *args, **kwargs)


  """
  Reuses previously computed samples, if possible
  """
  def __call__(self, table, samp_rate):
    f = Orange.data.filter.ValueFilterContinuous(
      position=self.SCORE_ID,
      oper=orange.ValueFilter.NotEqual,
      ref=-inf
    )
    c = Orange.data.filter.Values(
      domain=table.domain,
      conditions=[f],
      negate=True
    )
    scored = table.filter_ref(c)

    pop = len(table)
    samp_size = min(pop, samp_rate * pop + 1)
    if not samp_size:
      return table

    if len(scored) >= samp_size:
      # only use 0.5 from previously computed samples
      p1 = (1. * samp_size) / float(len(scored))
      indices2 = Orange.data.sample.SubsetIndices2(p0=p1)
      idxs = indices2(scored) if len(scored) > 1 else [0]
      scored = scored.select_ref(idxs, negate=True)

    samp_size -= len(scored)
    c.negate = False        
    unscored = table.filter_ref(c)        
    try:
      p0 = float(samp_size) / len(unscored)
      indices2 = Orange.data.sample.SubsetIndices2(p0=p0)
      idxs = indices2(unscored) if len(unscored) > 1 else [0]
      samples = unscored.select_ref(idxs, negate=True)
    except:
      pdb.set_trace()

    samples.extend(scored)
    return samples






import sys
import pdb
import random
sys.path.extend(('.', '..'))

from scorpion.misc.gensinglecluster import gen_points, gen_multi_outliers
from scorpion.util import get_logger

from common import *

_logger = get_logger()
random.seed(0)


def init_db(db):
  try:
    with db.begin() as conn:
        conn.execute("""create table stats (
    expid int,
    id serial,
    tstamp timestamp default current_timestamp,
    dataset text,
    klass varchar(128) null,
    cols text null,
    epsilon float null,
    c float null,
    lambda float null,
    cost float,
    acc float,
    prec float,
    recall float,
    f1 float,
    score float,
    isbest bool,
    ischeckpoint bool default FALSE,
    checkpoint float null,
    completed bool,
    rule text null,
    ids text null,
    notes text null,
    boundtype text null
            )""")
        conn.execute("""create table costs (
        id serial,
        sid int,
        name varchar(128),
        cost float)""")
  except:
      pass

  try:
    with db.begin() as conn:
      conn.execute(""" CREATE TABLE dataset_metadata (
        id serial,
        tablename text,
        npts int,
        ndim int,
        kdim int,
        volperc float,
        uh float,
        sh float,
        uo float,
        so float,
        boxes text) 
      """)
  except Exception as e:
    print e


def setup_synth_metadata(tablename, params, oboxes, sboxes):
  from scorpion.db import connect
  eng = connect('sigmod')
  db = eng.connect()
  val = json.dumps({
    'oboxes': oboxes,
    'sboxes': sboxes
  })

  q = "insert into dataset_metadata values(DEFAULT, %s)" % ', '.join(["%s"]*10)
  args = [tablename]
  args.extend(params)
  args.append(val)
  db.execute(q, args)
  db.close()
  eng.dispose()

def setup_synth_data(tablename, schema, pts):
  with file('/tmp/clusters.txt', 'w') as f:
      print >>f, '\t'.join(schema)
      for pt in pts:
          print >>f, '\t'.join(['%.4f']*len(pt)) % tuple(pt)
  os.system('importmydata.py %s sigmod /tmp/clusters.txt 2> /tmp/dbtruck.err > /tmp/dbtruck.log' % tablename)

def setup_sigmod(ndim, volperc):
  params = []
  #for kdim in xrange(1, ndim+1):
  kdim = ndim
  npts = 2000
  for uo in (30, 80):
    tablename = 'data_%d_%d_%d_0d%d_%duo'
    args = (ndim, kdim, npts, int(100.*volperc), uo)
    tablename = tablename % args
    params.append((tablename, (npts, ndim, kdim, volperc, 10, 10, uo, 10)))


  for tablename, args in params:
    obox, sbox, schema, pts = gen_points(*args)
    setup_synth_data(tablename, schema, pts)
    setup_synth_metadata(tablename, args, [obox], [sbox])

def setup_multicluster(ndim, volperc):
  print >>sys.stderr, "loading multi-cluster outliers"

  npts = 2000
  params = []
  for uo in (30, 80):
    tablename = 'data2clust_%d_%d_%d_vol%d_uo%d' 
    args = (ndim, ndim, npts, int(100*volperc), uo)
    tablename = tablename % args
    params.append((tablename, (2000, ndim, ndim, volperc, 10, 10, uo, 10)))
 
  for tablename, args in params:
    oboxes, sboxes, schema, pts = gen_multi_outliers(*args)
    setup_synth_data(tablename, schema, pts)
    setup_synth_metadata(tablename, args, oboxes, sboxes)

if __name__ == '__main__':
  from sqlalchemy import create_engine
  db = create_engine('postgresql://localhost/sigmod')
  init_db(db) 




  setup_sigmod(2, 0.5)
  setup_sigmod(3, 0.5)
  setup_sigmod(4, 0.5)
  setup_sigmod(5, 0.5)
  setup_multicluster(2, .15)
  setup_multicluster(3, .15)
  setup_multicluster(2, .2)
  setup_multicluster(3, .2)
  setup_multicluster(4, .15)
  setup_multicluster(4, .2)
  setup_sigmod(2, 0.01)
  setup_sigmod(3, 0.01)
  setup_sigmod(4, 0.01)
  setup_sigmod(5, 0.01)
  setup_sigmod(2, 0.1)
  setup_sigmod(3, 0.1)
  setup_sigmod(4, 0.1)
  setup_sigmod(5, 0.1)



  db.dispose()




import pdb
import sys
import heapq
import bisect
import numpy as np
sys.path.extend(['.', '..'])

from collections import deque
from itertools import chain

from ..util import *

inf = float('inf')
_logger = get_logger()


class CandidateQueue():
  def __init__(self, exp=3, highexp=None):
    self.q = []
    self.exp = exp
    self.highexp = highexp
    if not self.highexp:
      self.highexp = self.exp * 2

  def __len__(self):
    return len(self.q)

  def add(self, tup):
    """
    ARG: 
        tup: (score, estinf, (node, datas, srs, infs))
    """
    _logger.debug("%.4f\t%.4f\t%s" % (tup[0], tup[1], tup[2][0].rule))
    bisect.insort_left(self.q, tup)
    #self.q.append(tup)

  def next(self):
    if not self.q:
      return None
    
    first = self.q[0]
    _logger.debug("first: %.4f\t%.4f\t%s", 
      first[0], first[1], first[2][0].rule)

    if len(self.q) == 1:
      self.total_score = 0
      return self.q.pop()[2:]

    exp = self.exp
    if random.random() < .5:
      exp = self.highexp
    idx = int((1-np.random.power(exp)) * len(self.q))
    return self.q.pop(idx)[2:]



import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

from collections import deque
from itertools import chain
from multiprocessing import Process, Queue, Pool, Pipe
from Queue import Empty

from ..learners.cn2sd.rule import *
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *
from ..settings import *

from basic import Basic
from sampler import Sampler
from merger import Merger
from node import Node

INF = float('inf')
_logger = get_logger()


def partition_f(name, params, learner, tables, full_table, stats, (inq, outq)):
  try:
    partitioner = BDTTablesPartitioner(**params)
    partitioner.setup_tables(tables, full_table)
    partitioner.stats = stats
    gen = partitioner()
    start = time.time()
    while not partitioner.is_done:
      try: 
        (cmd, payload) = inq.get(False)
        if cmd == 'bound':
          bound = payload
          partitioner.update_inf_bound(bound)
        elif cmd == 'leaf':
          dicts = payload
          rules = [SDRule.from_json(d, full_table) for d in dicts]
          nodes = [Node(rule, None) for rule in rules]
          for node in nodes:
            partitioner.add_child(node)

      except Empty:
        pass

      pairs = []
      for node, isleaf in gen:
        rule = node.rule
        rule.quality = node.influence
        rule = rule.simplify(full_table, learner.cont_dists, learner.disc_dists)
        pairs.append((rule, isleaf))
        if len(pairs) >= 10: break

      if not pairs: 
        if not partitioner.is_done:
          _logger.debug("%s\tno nodes from generator but partitioner not done..."%name)
        continue

      bound = partitioner.get_inf_bound()

      dicts = [(r.to_json(), isleaf) for r, isleaf in pairs]
      #_logger.debug("%s\tsend %d rules\t%s" % (name, len(pairs), map(hash, map(str, dicts))))
      outq.put((dicts, bound))
      #_logger.debug("%s\tsent %s!" % (name, len(dicts)))
  except Exception as e:
    print e
    import traceback
    traceback.print_exc()

  _logger.debug("%s\tpartitioner DONE" % name)
  outq.put('done')
  outq.close()
  inq.close()



      








class BDTTablesPartitioner(Basic):

  def set_params(self, **kwargs):
      Basic.set_params(self, **kwargs)

      self.p = kwargs.get('p', 0.6)
      self.tau = kwargs.get('tau', [0.001, 0.09])
      self.epsilon = kwargs.get('epsilon', 0.005)
      self.partition_min_pts = 2
      self.min_pts = 5
      self.SCORE_ID = kwargs['SCORE_ID']
      self.inf_bounds = None 
      self.min_improvement = kwargs.get('min_improvement', .01)
      self.err_funcs = kwargs.get('err_funcs', None)


      self.max_wait = kwargs.get('max_wait', 2*60*60)
      self.start_time = None

      self.sampler = Sampler(self.SCORE_ID)
      self.seen = set()
      self.is_done = False

      self.roots = []


      if self.err_funcs is None:
        raise RuntimeError("errfuncs is none")


  @instrument
  def setup_tables(self, tables, merged):
      self.merged_table = merged
      self.tables = tables

      for ef, t in zip(self.err_funcs, self.tables):
        ef.setup(t)

      self.sampler = Sampler(self.SCORE_ID)
      self.samp_rates = [best_sample_size(len(t), self.epsilon)/(float(len(t))+1) for t in self.tables]

      if self.inf_bounds is None:
        self.inf_bounds = [[INF, -INF] for table in tables]

      # attributes to partition upon
      self.cont_attrs = [attr.name for attr in merged.domain if attr.name in self.cols and attr.var_type != Orange.feature.Type.Discrete]
      self.dist_attrs = [attr.name for attr in merged.domain if attr.name in self.cols and attr.var_type == Orange.feature.Type.Discrete]

      # remove undesirable columns 
      self.cont_attrs = filter(lambda c: c in self.cols, self.cont_attrs)
      self.dist_attrs = filter(lambda c: c in self.cols, self.dist_attrs)

  def __call__(self, tables=None, full_table=None, root=None, **kwargs):
    if tables and full_table:
      self.setup_tables(tables, full_table)

    self.seen = set()
    if not root:
      root = Node(SDRule(self.merged_table, None))
    self.roots.extend(root.leaves)

    samples = [self.sample(t, sr) for t, sr in zip(self.tables, self.samp_rates)]
    f = lambda (idx, samps): self.compute_infs(idx, samps)
    all_infs = map(f, enumerate(samples))

    while self.roots:
      leaf = self.roots.pop()
      parent = leaf.parent
      leaf.parent = None
      for pair in self.grow(leaf, samples, self.samp_rates, all_infs):
        n, isleaf = pair
        _logger.debug("yield:\t%s", str(n))
        yield pair
      leaf.parent = parent

    self.is_done = True

  def add_root(self, node):
    self.roots.extend(node.leaves())

  def get_inf_bound(self):
    bound = [INF, -INF]
    for inf_bound in self.inf_bounds:
      bound = r_union(bound, inf_bound)
    return bound

  def update_inf_bound(self, bound):
    for idx, inf_bound in enumerate(self.inf_bounds):
      self.inf_bounds[idx] = r_union(bound, inf_bound)

  @instrument
  def sample(self, data, samp_rate):
      return self.sampler(data, samp_rate)

  def should_idx_stop(self, args):
    idx, infs = args
    if len(infs) <= self.partition_min_pts:
        return True

    infmax = max(infs)
    thresh = self.compute_threshold(infmax, idx)
    std = np.std(infs)
    maxv, minv = max(infs), min(infs)
    return std < thresh #and maxv - minv < thresh

  def print_status(self, rule, datas, sample_infs):
    bools = map(self.should_idx_stop, enumerate(sample_infs))
    perc_passed = np.mean(map(float, bools))

    maxstd, maxrange, minthresh = -INF, -INF, INF
    maxbound = -INF
    perc = -INF
    maxidx = -1
    for idx, infs in enumerate(sample_infs):
      if not infs: continue
      std = np.std(infs)
      if std > maxstd:
        maxstd = std
        maxidx = idx
        maxbound = self.inf_bounds[idx]
        if maxbound:
          perc = (max(infs) - self.inf_bounds[idx][0]) / (maxbound[1]-maxbound[0])
        else:
          perc = -INF
      maxrange = max(maxrange, max(infs)-min(infs))
      minthresh = min(minthresh, self.compute_threshold(max(infs), idx))
    npts = sum(map(len, datas))

    fmt = "status:\t%.3f\tid(%d)\tn(%d)\tstd(%.4f)\tperc(%.4f)\tbnd(%.3f, %.3f)\tthrsh(%.4f)\t%s"
    try:
      _logger.debug(fmt,
          perc_passed,
          maxidx,
          npts,
          maxstd,
          perc,
          maxbound[0],
          maxbound[1],
          minthresh,
          str(rule))
    except:
      pdb.set_trace()


  def should_stop(self, sample_infs): 
    bools = map(self.should_idx_stop, enumerate(sample_infs))
    if sum(map(len, sample_infs)) < self.min_pts: 
      return True
    return reduce(and_, bools)


  def influence(self, row, idx):
    if row[self.SCORE_ID].value == -INF:
      influence = self.err_funcs[idx]((row,))
      row[self.SCORE_ID] = influence
      self.inf_bounds[idx][0] = min(influence, self.inf_bounds[idx][0])
      self.inf_bounds[idx][1] = max(influence, self.inf_bounds[idx][1])
    return row[self.SCORE_ID].value

  def compute_threshold(self, infmax, idx):
      infl, infu = tuple(self.inf_bounds[idx])
      tau, p = self.tau, self.p
      inflection = p*(infu-infl) + infl
      s = (tau[0] - tau[1]) / (infu - inflection)
      w = tau[0] + s*(infmax - infu)
      w = min(tau[1], w)
      ret = w * (infu - infl)       
      if ret == -INF:
          raise RuntimeError()
      if ret < 0:
        pdb.set_trace()
      return ret


  def compute_infs(self, idx, samples):
    return [self.influence(r, idx) for r in samples]

  def estimate_inf(self, sample_infs):
    means = map(np.mean, filter(bool, sample_infs))
    if means:
      return np.mean(means)
    return -INF


  @instrument
  def databyrule2infs(self, rules, datas):
    """
    compute list of influence values for each combination of rule x data
    @param rules list of rule
    @param datas list of data
    @return 
    """
    data2infs = defaultdict(list)
    rule2infs = defaultdict(list)
    rule2datas = defaultdict(list)
    for idx, data in enumerate(datas):
      for rule in rules:
        filtered = rule.filter(data)
        infs = self.compute_infs(idx, filtered)
        rule2infs[rule].append(infs)
        data2infs[idx].append(infs)
        rule2datas[rule].append(filtered)
    return data2infs, rule2infs, rule2datas


  @instrument
  def get_scores(self, rules, samples):
    sample2infs, rule2infs, _ = self.databyrule2infs(rules, samples)
    scores = []

    for idx in xrange(len(samples)):
      allinfs = sample2infs[idx]
      score = self.get_score_for_infs([idx]*len(allinfs), allinfs)
      scores.append(score)

    for r in rules:
      allinfs = rule2infs[r]
      r.quality = self.estimate_inf(allinfs)

    scores = filter(lambda s: s!=-INF, scores)
    return scores

  @instrument
  def get_score_for_infs(self, idxs, samp_infs):
      scores, counts = [], []
      for idx, infs in zip(idxs, samp_infs):
        if not len(infs): continue
        thresh = self.compute_threshold(max(infs), idx)
        bounds = self.inf_bounds[idx]
        inf_range = bounds[1] - bounds[0]
        if not inf_range:
          scores.append(0)
          counts.append(len(infs))
        else:
          std = np.std(infs)
          #scores.append(((thresh - bounds[0]) / inf_range) * (std - thresh))
          scores.append((1. - (thresh / inf_range)) * (std - thresh))
          counts.append(len(infs))
      if scores:
        return np.mean(scores)
      return -INF

  def merge_scores(self, scores):
      if scores:
        return np.percentile(scores, 75) #10)#75) #XXX: hack
      return -INF

  def adjust_score(self, score, node, attr, rules):
    # penalize excessive splitting along a single dimension if it is not helping
    if attr.var_type == Orange.feature.Type.Discrete:
      if attr == node.prev_attr:
        score = score - (0.15) * abs(score)
    else:
      if attr == node.prev_attr:
        #score = score + (0.01) * abs(score)
        if False and self.skinny_penalty(rules):
          score = score + (0.6) * abs(score)
    return score


  def get_states(self, tables):#node):
      # find tuples in each table that is closest to the average influence
      states = []
      all_infs = []
      for idx, table in enumerate(tables):
        if len(table) == 0:
          states.append(None)
          continue
        infs = self.compute_infs(idx, table)
        all_infs.append(infs)
        min_idx = min(range(len(table)), key=lambda i: infs[i])
        min_tup = table[min_idx]
        state = self.err_funcs[idx].state((min_tup,))
        states.append(state)
      return states
  

  @instrument
  def skinny_penalty(self, rules):
    for rule in rules:
      edges = []
      for c in rule.filter.conditions:
          attr = self.merged_table.domain[c.position]
          if attr.var_type == Orange.feature.Type.Discrete:
              continue
          edges.append(c.max - c.min)
      if len(edges) > 1:
          volume = reduce(mul, edges)
          mean_edge = sum(edges) / float(len(edges))
          max_vol = mean_edge ** len(edges)
          perc = (volume / max_vol) ** (1./len(edges))
          if perc < 0.05:
              return True
          return (1. - perc) * 1.5
      return False

      
  def time_exceeded(self):
    return (
        self.start_time is not None and
        self.max_wait is not None and
        (time.time() - self.start_time) >= self.max_wait
        )
      


  def grow(self, node, tables, samp_rates, sample_infs=None):
    if self.time_exceeded():
      _logger.debug("time exceeded %.2f > %d", (time.time()-self.start_time), self.max_wait)
      yield (node, False)
      return

    if node.rule in self.seen:
      _logger.debug("rule seen %d\t%s", hash(node.rule), node.rule)
      yield (node, False)
      return
    self.seen.add(node.rule)

    if self.start_time is None and node.depth >= 1:
      self.start_time = time.time()

    rule = node.rule
    datas = tables
    if not sample_infs:
      datas = map(rule.filter_table, tables)
    node.cards = map(len, datas)
    node.n = sum(node.cards)

    if node.n == 0:
      yield (node, False)
      return

    #
    # Precompute influences and scores
    #

    samples = datas
    if sample_infs is None:
      f = lambda (idx, samps): self.compute_infs(idx, samps)
      samples = [self.sample(*pair) for pair in zip(datas, samp_rates)]
      sample_infs = map(f, enumerate(samples))

    curscore = self.get_score_for_infs(range(len(sample_infs)), sample_infs)
    est_inf = self.estimate_inf(sample_infs)
    node.set_score(est_inf)

    if node.parent:
      self.print_status(rule, datas, sample_infs)
      if self.should_stop(sample_infs):
        node.states = self.get_states(datas)
        yield (node, True)
        return


    if self.time_exceeded():
      _logger.debug("time exceeded %.2f > %d", (time.time()-self.start_time), self.max_wait)
      yield (node, False)
      return 


    #
    # compute scores for each attribute to split on
    #
    attr_scores = []
    for attr, new_rules in self.child_rules(rule):
      if not new_rules: continue
      scores = self.get_scores(new_rules, samples)
      score = self.merge_scores(scores)
      score = self.adjust_score(score, node, attr, new_rules)
      _logger.debug("score:\t%.4f\t%s\t%s", score, attr.name[:6], new_rules[0])
      if score == -INF: continue
      attr_scores.append((attr, new_rules, score, scores))


    if not attr_scores:
      node.states = self.get_states(datas)
      yield (node, True)
      return

    attr_scores.sort(key=lambda p: p[-2])

    attr, new_rules, score, scores = attr_scores[0]
    node.score = min(scores) 
    minscore = curscore - abs(curscore) * self.min_improvement
    if node.score >= minscore and minscore != -INF:
      _logger.debug("bdt:  \tscore didn't improve\t%.7f >= %.7f", min(scores), minscore)
      yield (node, True)
      return

    if node.score <= curscore - abs(curscore) * 0.05:
      _logger.debug("bdt:  \tbig improvement\t%s", str(new_rules[0]))
      yield (node, False)

    ncands = max(1, 2 - node.depth)
    for attr, new_rules, score, scores in attr_scores[:ncands]:
      data2infs, rule2infs, rule2datas = self.databyrule2infs(new_rules, datas)
      #new_srses = self.update_sample_rates(new_rules, data2infs, samp_rates)
      new_srses = [samp_rates] * len(new_rules)
      new_pairs = zip(new_rules, new_srses)
      new_pairs.sort(key=lambda (new_r, new_s): new_r.quality, reverse=True)

      for new_rule, new_srs in new_pairs:
        child = Node(new_rule)
        child.prev_attr = attr
        child.parent = node
        node.add_child(child)

        args = (child, rule2datas[new_rule], new_srs, rule2infs[new_rule])
        #self.candidates.add((score, new_rule.quality, args))
        for pair in self.grow(*args):
          yield pair

    yield (node, False)


  @instrument
  def child_rules(self, rule):
      next_rules = defaultdict(list)

      if self.cont_attrs:
          refiner = BeamRefiner(attrs=self.cont_attrs, fanout=2)
          for attr, new_rule in refiner(rule):
              next_rules[attr].append(new_rule)
      if self.dist_attrs:
          refiner = BeamRefiner(attrs=self.dist_attrs, fanout=5)
          for attr, new_rule in refiner(rule):
              next_rules[attr].append(new_rule)
      return next_rules.items()



  @instrument
  def update_sample_rates(self, rules, data2infs, srs):
    """return list where each element are the sample rates for a single rule across the tables"""
    srs_by_table = [[0]*len(srs) for i in data2infs]
    for idx in data2infs:
      sr = srs[idx]
      all_infs = data2infs[idx]

      if not sr: continue
      if not sum(map(len, all_infs)): continue

      new_srs = self.update_sample_rates_helper(all_infs, sr, idx)
      srs_by_table[idx] = new_srs
    return zip(*srs_by_table)

  def update_sample_rates_helper(self, all_infs, samp_rate, idx):
      influences, counts = [], []
      for infs in all_infs:
        influence = sum(np.array(infs)-self.inf_bounds[idx][0])
        influences.append(influence)
        counts.append(len(infs) + 1.)

      total_inf = sum(influences)
      total_count = sum(counts)
      if not total_inf:
        return [0]*len(all_infs)
      samp_rates = []
      nsamples = total_count * samp_rate
      for influence, count in zip(influences, counts):
        infr = influence / total_inf
        sub_samples = infr * nsamples
        nsr = sub_samples / count
        nsr = max(0, min(1., nsr))
        samp_rates.append(nsr)

      return samp_rates







import time
import pdb
import sys
import Orange
import orange
import heapq
import bisect
import numpy as np
sys.path.extend(['.', '..'])

from collections import deque
from itertools import chain

from rtree.index import Index as RTree
from rtree.index import Property as RProp

from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *
from ..settings import *

from bdtq import *
from node import Node
from basic import Basic
from sampler import Sampler
from merger import Merger
from node import Node

inf = float('inf')
_logger = get_logger()



class BDTTablesPartitioner(Basic):

    def set_params(self, **kwargs):
        Basic.set_params(self, **kwargs)

        self.p = kwargs.get('p', 0.6)
        self.tau = kwargs.get('tau', [0.001, 0.15])
        self.epsilon = kwargs.get('epsilon', 0.005)
        self.min_pts = 5
        self.SCORE_ID = kwargs['SCORE_ID']
        self.inf_bounds = None 
        self.min_improvement = kwargs.get('min_improvement', .01)

        self.candidates = CandidateQueue()

        # kill off partitioning after max_wait exceeded
        self.max_wait = kwargs.get('max_wait', 2*60*60) # 2 hours
        self.start_time = None

        self.sampler = Sampler(self.SCORE_ID)


    def setup_tables(self, tables, merged):
        self.merged_table = merged
        self.tables = tables
        self.err_funcs = [self.err_func.clone() for t in tables]

        for ef, t in zip(self.err_funcs, self.tables):
            ef.setup(t)

        self.sampler = Sampler(self.SCORE_ID)
        self.samp_rates = [best_sample_size(len(t), self.epsilon)/(float(len(t))+1) for t in self.tables]

        if not self.inf_bounds or len(self.inf_bounds) != len(tables):
          self.inf_bounds = [[inf, -inf] for table in tables]


    def __call__(self, tables, full_table, root=None, **kwargs):
        self.setup_tables(tables, full_table)
        if not root:
          root = Node(SDRule(full_table, None))
        self.root = root

        for leaf in root.leaves:
          parent = leaf.parent
          leaf.parent = None
          self.grow(leaf, self.tables, self.samp_rates)
          leaf.parent = parent

        return self.root.nodes

    def sample(self, data, samp_rate):
        return self.sampler(data, samp_rate)

    def should_idx_stop(self, args):
      idx, infs = args
      if len(infs) < self.min_pts:
          return True

      infmax = max(infs)
      thresh = self.compute_threshold(infmax, idx)
      std = np.std(infs)
      maxv, minv = max(infs), min(infs)
      return std < thresh and maxv - minv < thresh

    def print_status(self, rule, datas, sample_infs):
      bools = map(self.should_idx_stop, enumerate(sample_infs))
      perc_passed = np.mean(map(float, bools))

      all_infs = filter(bool, sample_infs)
      maxstd = max(map(np.std, all_infs))
      maxrange = max([max(infs)-min(infs) for infs in all_infs])
      threshes = [self.compute_threshold(max(infs), idx) for idx, infs in enumerate(all_infs)]
      npts = sum(map(len, datas)) if datas else 0
      _logger.debug("%.3f\tstd(%.4f)\trange(%.4f)\tthresh(%.4f)\t%s",
          perc_passed,
          maxstd,
          maxrange,
          min(threshes),
          str(rule))

 
    def should_stop(self, sample_infs): 
      bools = map(self.should_idx_stop, enumerate(sample_infs))
      return reduce(and_, bools)

    def influence(self, row, idx):
       if row[self.SCORE_ID].value == -inf:
          influence = self.err_funcs[idx]((row,))
          row[self.SCORE_ID] = influence
          self.inf_bounds[idx][0] = min(influence, self.inf_bounds[idx][0])
          self.inf_bounds[idx][1] = max(influence, self.inf_bounds[idx][1])
       return row[self.SCORE_ID].value


    def compute_infs(self, idx, data):
      return [self.influence(r,idx) for r in data]

    def merge_scores(self, scores):
        if scores:
          return float(np.percentile(scores, 75))
        return -inf

    def compute_threshold(self, infmax, idx):
        infl, infu = tuple(self.inf_bounds[idx])
        tau, p = self.tau, self.p
        s = (tau[0] - tau[1]) / ((1-p)*infu - p * infl)
        w = tau[0] + s*(infmax - infu)
        w = min(tau[1], w)
        ret = w * (infu - infl)       
        if ret == -inf:
          pdb.set_trace()
          raise RuntimeError()
        return ret


    def estimate_inf(self, sample_infs):
        return np.mean(map(np.mean, filter(bool, sample_infs)))

    def databyrule2infs(self, rules, datas):
      """
      compute list of influence values for each combination of rule x data
      @param rules list of rule
      @param datas list of data
      @return 
      """
      data2infs = defaultdict(list)
      rule2infs = defaultdict(list)
      rule2datas = defaultdict(list)
      for idx, data in enumerate(datas):
        for rule in rules:
          filtered = rule.filter(data)
          infs = self.compute_infs(idx, filtered)
          rule2infs[rule].append(infs)
          data2infs[idx].append(infs)
          rule2datas[rule].append(filtered)
      return data2infs, rule2infs, rule2datas

    def get_scores(self, rules, samples):
      sample2infs, rule2infs, _ = self.databyrule2infs(rules, samples)
      scores = []

      for idx in xrange(len(samples)):
        allinfs = sample2infs[idx]
        score = self.get_score_for_infs([idx]*len(allinfs), allinfs)
        scores.append(score)

      scores = filter(lambda s: s!=-inf, scores)
      return scores

    def get_score_for_infs(self, idxs, samp_infs):
        scores, counts = [], []
        for idx, infs in zip(idxs, samp_infs):
          if not len(infs): continue
          thresh = self.compute_threshold(max(infs), idx)
          bounds = self.inf_bounds[idx]
          inf_range = bounds[1] - bounds[0]
          if not inf_range:
            scores.append(0)
            counts.append(n)
          else:
            std = np.std(infs)
            scores.append(((thresh - bounds[0]) / inf_range) * (std - thresh))
            counts.append(len(infs))
        if scores:
          return np.mean(scores)
        return -inf

    def merge_scores(self, scores):
        if scores:
          return np.percentile(scores, 75)
        return -inf


    def get_states(self, tables):#node):
        #tables = map(node.rule,self.tables)

        # find tuples in each table that is closest to the average influence
        all_infs = []
        for idx, table in enumerate(tables):
            infs = [self.influence(row, idx) for row in table]
            all_infs.append(infs)
        states = []

        for idx, t, infs in zip(xrange(len(tables)), tables, all_infs):
            if infs:
                avg = np.mean(infs)
                min_tup = min(t, key=lambda row: self.influence(row, idx))
                state = self.err_funcs[idx].state((min_tup,))
                states.append(state)
            else:
                states.append(None)

        return states
    
    def adjust_score(self, score, node, attr, rules):
      # penalize excessive splitting along a single dimension if it is not helping
      if attr.var_type == Orange.feature.Type.Discrete:
        score = score - (0.15) * abs(score)
      else:
        if attr == node.prev_attr:
          score = score + (0.01) * abs(score)
          if False and self.skinny_penalty(rules):
            score = score + (0.6) * abs(score)
      return score


    def skinny_penalty(self, rules):
      for rule in rules:
        edges = []
        for c in rule.filter.conditions:
            attr = self.merged_table.domain[c.position]
            if attr.var_type == Orange.feature.Type.Discrete:
                continue
            edges.append(c.max - c.min)
        if len(edges) > 1:
            volume = reduce(mul, edges)
            mean_edge = sum(edges) / float(len(edges))
            max_vol = mean_edge ** len(edges)
            perc = (volume / max_vol) ** (1./len(edges))
            if perc < 0.05:
                return True
            return (1. - perc) * 1.5
        return False
        
    def time_exceeded(self):
      return ( 
          self.start_time is not None and
          self.max_wait is not None and 
          time.time() - self.start_time >= self.max_wait
      )

    def adjust_score(self, score, node, attr, rules):
      # prefer discrete attributes 
      # XXX: (based on cardinality???)
      if attr.var_type == Orange.feature.Type.Discrete:
        score -= (0.151) * abs(score)

      # penalize excessive splitting along a single dimension if it is not helping
      else:
        if False and attr == node.prev_attr:
          #score = score + (0.01) * abs(score)
          penalty = self.skinny_penalty(rules)
          score += abs((0.25) * penalty * score)

      return score

    def grow(self, node, tables, samp_rates, sample_infs=None):
        if self.time_exceeded():
          _logger.debug("time exceeded %.3f > %.3f", time.time()-self.start_time, self.max_wait)
          return node

        rule = node.rule
        data = rule.examples
        datas = map(rule.filter_table, tables)
        samples = [self.sample(*pair) for pair in zip(datas, samp_rates)]
        node.cards = map(len, datas)
        node.n = sum(node.cards)

        if node.n == 0:
          return node

        #if sample_infs is None:
        f = lambda (idx, samps): self.compute_infs(idx, samps)
        sample_infs = map(f, enumerate(samples))
        curscore = self.get_score_for_infs(range(len(samples)), sample_infs)
        est_inf = self.estimate_inf(sample_infs)
        node.set_score(est_inf)

        if node.parent:
          self.print_status(rule, datas, sample_infs)
          if self.should_stop(sample_infs):
            node.states = self.get_states(datas)
            return node


        attr_scores = []
        for attr, new_rules in self.child_rules(rule):
          scores = self.get_scores(new_rules, samples)
          score = self.merge_scores(scores)
          score = self.adjust_score(score, node, attr, new_rules)
          if score == -inf: continue
          attr_scores.append((attr, new_rules, score, scores))

        if not self.start_time: self.start_time = time.time()

        if not attr_scores:
          node.states = self.get_states(datas)
          _logger.debug("no attrscores")
          return node

        attr, new_rules, score, scores = min(attr_scores, key=lambda p: p[-2])

        node.score = min(scores) 
        minscore = curscore - abs(curscore) * self.min_improvement
        if node.score >= minscore and minscore != -inf:
          _logger.debug("score didn't improve\t%.7f >= %.7f", min(scores), minscore)
          return node


        #data2infs, rule2infs, rule2datas = self.databyrule2infs(new_rules, datas)
        #new_srses2 = self.update_sample_rates2(new_rules, data2infs, samp_rates)
        new_srses = self.update_sample_rates(new_rules, datas, samp_rates)
        #for newsrs, newsrs2 in zip(new_srses, new_srses2):
          #if tuple(newsrs) != tuple(newsrs2):
            #pdb.set_trace()


        for new_rule, new_srs in zip(new_rules, new_srses):
          child = Node(new_rule)
          child.prev_attr = attr
          child.parent = node

          #self.grow(child, rule2datas[new_rule], new_srs)#, rule2infs[new_rule])
          self.grow(child, datas, new_srs)

          if child and child.n and child.influence != -inf:
            node.add_child(child)

        return node


    def child_rules(self, rule, attrs=None):
        attrs = attrs or self.cols
        next_rules = defaultdict(list)
        cont_attrs = [attr.name for attr in self.merged_table.domain if attr.name in attrs and attr.var_type != Orange.feature.Type.Discrete]
        dist_attrs = [attr.name for attr in self.merged_table.domain if attr.name in attrs and attr.var_type == Orange.feature.Type.Discrete]

        if cont_attrs:
            refiner = BeamRefiner(attrs=cont_attrs, fanout=2)
            for attr, new_rule in refiner(rule):
                next_rules[attr].append(new_rule)
        if dist_attrs:
            refiner = BeamRefiner(attrs=dist_attrs, fanout=8)
            for attr, new_rule in refiner(rule):
                next_rules[attr].append(new_rule)
        return next_rules.items()


    def update_sample_rates2(self, rules, data2infs, srs):
      """return list where each element are the sample rates for a single rule across the tables"""
      srs_by_table = [[0]*len(srs) for i in data2infs]
      for idx in data2infs:
        sr = srs[idx]
        all_infs = data2infs[idx]

        if not sr: continue
        if not sum(map(len, all_infs)): continue

        new_srs = self.update_sample_rates_helper2(all_infs, sr, idx)
        srs_by_table[idx] = new_srs
      return zip(*srs_by_table)

    def update_sample_rates_helper2(self, all_infs, samp_rate, idx):
        influences, counts = [], []
        for infs in all_infs:
          inf = sum(np.array(infs)-self.inf_bounds[idx][0])
          influences.append(inf)
          counts.append(len(infs) + 1.)

        total_inf = sum(influences)
        total_count = sum(counts)
        if not total_inf:
          return [0]*len(all_infs)
        samp_rates = []
        nsamples = total_count * samp_rate
        for influence, count in zip(influences, counts):
          infr = influence / total_inf
          sub_samples = infr * nsamples
          nsr = sub_samples / count
          nsr = max(0, min(1., nsr))
          samp_rates.append(nsr)

        return samp_rates



    def update_sample_rates(self, rules, tables, srs):
        srs_by_table = [[0]*len(srs) for i in tables]
        for idx, (t, samp_rate) in enumerate(zip(tables, srs)):
            if not samp_rate:
                continue
            new_tables = [r.filter_table(t) for r in rules]
            if not sum(map(len, new_tables)):
                continue
            new_samp_rates = self.update_sample_rates_helper(new_tables, samp_rate, idx)
            srs_by_table[idx] = new_samp_rates
        return zip(*srs_by_table)

    def update_sample_rates_helper(self, datas, samp_rate, idx):
        influences, counts = [], []
        f = lambda row: self.influence(row, idx)
        for data in datas:
            influence = map(f, data)
            influence = sum(i - self.inf_bounds[idx][0] for i in influence)
            influences.append(influence)
            counts.append(len(data)+1.)

        total_inf = sum(influences)
        total_count = sum(counts)
        if not total_inf:
            return [0]*len(datas)
        samp_rates = []
        nsamples = total_count * samp_rate
        for influence, count in zip(influences, counts):
            infr = influence / total_inf
            sub_samples = infr * nsamples
            nsr = sub_samples / count
            nsr = max(0, min(1., nsr))
            samp_rates.append(nsr)

        return samp_rates



#
# Validates that table doesn't contain any nulls
# and optionally replaces null values with sane defaults
#
import click
from sqlalchemy import create_engine
from datetime import date



def get_cols(db, tablename):
  q = """
  select attname from pg_class, pg_attribute 
  where relname = %s and attrelid = pg_class.oid and attnum > 0 and attisdropped = false;
  """
  ret = []
  for (attr,) in db.execute(q, tablename):
    ret.append(attr)
  return ret


def fix_typ(db, tablename, col, count, typ, bfix):
  vals = []
  if 'char' in typ:
    t = str
    vals = ['', 'NULL', '_NULL_']
  elif 'int' in typ:
    t = int
    vals = [0, -1]
  elif 'float' in typ:
    t = float
    vals = [0, -1]
  elif 'date' in typ:
    vals = [date(1000, 1, 1)]
    t = date
  else:
    vals = []
    print "\t%s\t%s\tno defaults" % (typ, col)

  if 'char' in typ or 'text' in typ:
    print "removing single quotes"
    q = """
    UPDATE %s 
    SET %s = replace(%s, e'\\'', '') 
    WHERE substring(%s from e'\\'') is not null;
    """ % (tablename, col, col, col)
    print q
    if bfix:
      db.execute(q)
      try:
        db.commit()
      except Exception as e:
        print "\t%s" % e
      print "\tsuccess"

  if count == 0: return

  print "fixing %s as %s" % (col, typ)
  for val in vals:
    q = """
    SELECT count(*) from %s where %s = %%s
    """ % (tablename, col)
    nconflicts = db.execute(q, val).fetchone()[0]
    print "\tconflicts: %s\t%s" % (nconflicts, val)
    
    if nconflicts is 0:
      print "\tupdating values to %s" % val
      q = """
      UPDATE %s SET %s = %%s WHERE %s is null
      """ % (tablename, col, col)
      print q
      if bfix:
        db.execute(q, val)
        try:
          db.commit()
        except Exception as e:
          print "\t%s" % e
          pass
        print "\tsuccess"
        break





def check_and_fix(db, tablename, col, bfix):
  q = "SELECT count(distinct %s) from %s where %s is not null" % (col, tablename, col)
  q = "SELECT count(*) from %s where %s is null" % (tablename, col)
  row = db.execute(q).fetchone()
  count = row[0]

  q = """SELECT pg_type.typname FROM pg_attribute, pg_class, pg_type where 
    relname = %%s and pg_class.oid = pg_attribute.attrelid and attname = '%s' and
    pg_type.oid = atttypid"""
  row = db.execute(q % col, tablename).fetchone()
  typ = row[0]
  print "%s\t%s\t%s" % (count, typ, col)

  fix_typ(db, tablename, col, count, typ, bfix)


def get_tables(db):
  q = "select tablename from pg_tables where schemaname = 'public'"
  return [str(table) for (table,) in db.execute(q).fetchall()]



@click.command()
@click.option('--bfix', is_flag=True, help="apply fixes to database.  Not setting this is same as a dryrun")
@click.option('--noask', is_flag=True, help="skip last ditch prompt when --fix is set")
@click.argument('dbname')
@click.argument('tables', nargs=-1)
def run(bfix, noask, dbname, tables):
  if bfix and not noask:
    v = raw_input("going to modify database!  enter 'x' to abort: ")
    if 'x' in v:
      print "aborting"
      exit()


  db = create_engine("postgresql://localhost/%s" % dbname)

  if not tables:
    tables = get_tables(db)

  if not bfix:
    print "Dry Run"

  print dbname
  print tables

  print "count\ttype\tcol"
  print "-------------------"
  for table in tables:
    cols = get_cols(db, table)
    for col in cols:
      check_and_fix(db, table, col, bfix)


if __name__ == '__main__':
  run()



import logging
import sys
import random
import math
import matplotlib
import numpy as np


def wmean(vals, weights=None):
    if weights is None: return np.mean(vals)
    vals = np.array(vals)
    weights = np.array(weights).astype(float)
    return (vals * weights).sum() / weights.sum()


def wstd(vals, weights=None):
    """
    compute a weighted standard deviation
    """
    mean = wmean(vals, weights)
    vals = np.array(vals).astype(float)
    weights = np.array(weights).astype(float)    
    weights /= weights.sum()

    top = (weights * ((vals - mean)**2)).sum()
    bot = weights.sum()
    return math.sqrt(top / bot)
    


def max_prob(u, s):
    vs = []
    for i in xrange(1000):
        a = random.gauss(u,s)
        if a >= u:
            vs.append(a)
    return np.mean(vs)

def order(u1, s1, u2, s2):
    """
    -1 if u1 < u2
    1  if u1 > u2 
    """
    vs = []
    for i in xrange(1000):
        a,b = random.gauss(u1,s1), random.gauss(u2,s2)
        if a<u1 or b<u2:
            continue
        vs.append( a - b )
    return np.mean(vs)

def gauss(u, s):
    """
    returns a gaussian function defined by mean and standard deviation
    @param u mean
    @param s standard deviation
    """
    a,b,c = 1./ (s * math.sqrt(2*math.pi)), u, s
    f = lambda x: (a * math.exp(- (x-b)**2 / (2 * c**2) )) 
    return f
    

def gaussorder(u1, s1, u2, s2, nsteps=20.):
    f1, f2 = gauss(u1, s1), gauss(u2, s2)

    minv, maxv = max(u1, u2), max(u1 + s1*5, u2 + s2*5)
    v, inc = minv, (maxv-minv) / float(nsteps)
    vs = []
    while v <= maxv:
        a = f1(v) * v if v > u1 else 0.
        b = f2(v) * v if v > u2 else 0.
        vs.append( a - b )
        v += inc
    return np.mean(vs)
    


def welchs_ttest(n1, mean1, sem1, n2, mean2, sem2, alpha):
    svm1 = sem1**2 * n1
    svm2 = sem2**2 * n2
    if svm1 == 0 and svm2 == 0:
        return 1.

    t_s_prime = (mean1 - mean2)/math.sqrt(svm1/n1+svm2/n2)
    return t_s_prime
    t_alpha_df1 = scipy.stats.t.ppf(1-alpha/2, n1 - 1)
    t_alpha_df2 = scipy.stats.t.ppf(1-alpha/2, n2 - 1)
    t_alpha_prime = (t_alpha_df1 * sem1**2 + t_alpha_df2 * sem2**2) / \
    (sem1**2 + sem2**2)
    return abs(t_s_prime) > t_alpha_prime, t_s_prime, t_alpha_prime



def prob_no_error(pop, errprob, npts):
    """
    computers the probability from a population with error probability `errprob` that
    a random sample of `npts` points will not contain any error points

    @param pop population size
    @param errprob probability of an error point (e.g., user specified lower bound)
    @param npts sample size
    """
    def choose(n, k):
        n,k = int(n), int(k)
        v = 0.
        if k > n/2:
            v += sum(map(math.log, xrange(k,n+1)))
            v -= sum(map(math.log, xrange(1,n-k+1)))
        else:
            v += sum(map(math.log, xrange(n-k,n+1)))
            v -= sum(map(math.log, xrange(1,k+1))) 
        return v
    
    # c(pop*(1-errprob), npts) / c(pop, npts)
    c1 =  choose(pop*(1-errprob), npts)
    c2 = choose(pop, npts)
    return math.exp(c1 - c2)

    
    #import scipy as sp
    #c = sp.misc.comb
    #nerr = pop * errprob
    e = int(math.ceil(errprob * pop))  # number of error points that is in the sample
    # initial equation.  Has numerical overflow issues
    # prob = c(npts, e) * (errprob ** e) * ((1.-errprob) ** (npts - e))
    

    # npts choose e = npts! / (e! (npts-e)!)
    bce = 0
    for i in xrange(1, npts):
        bce += math.log(i)
    for i in xrange(1, e):
        bce -= math.log(i)
    for i in xrange(1, npts - e):
        bce -= math.log(i)


    prob = 0.
    # errprob ** e    
    prob += e * math.log(errprob)
    # (1 - errprob) ** (npts - e)
    prob += (npts - e) * math.log(1. - errprob)

    prob = bce + prob
    prob = math.pow(math.e, prob)
    return prob

def best_sample_size(pop, errprob, confidence=0.95):
    """
    given a population and an error probability, computes the the minimum sample size `s` such
    that with 95% confidence, `s` will contain at least one error point
    """
    sample_size, best_prob = None, None
    threshold = max(0, 1. - confidence)
    mins, maxs = 1, pop

    while maxs - mins > 20:
        size = max(1, int((maxs + mins) / 2.))
        #print size, '\t', prob_no_error(pop, errprob, size)
        good = prob_no_error(pop, errprob, size) < threshold
        if good:
            # either this is the best, or we should update the ranges and
            # look again
            if prob_no_error(pop, errprob, size-1) < threshold:
                maxs = size
                continue
            else:
                return size
        else:
            mins = size+1
    for size in xrange(mins, maxs+1, 1):
        if prob_no_error(pop, errprob, size) < threshold:
            return size
    return pop


def sample_size(moe=0.1, pop=10, zval=2.58):
    """
    sample size based on estimator closed form solutions
    @param moe margin of error
    @param pop population size (size of partition)
    @param zval confidence interval (default: 99%)  95% is 1.96
    """
    ss = ((zval**2) * 0.25) / (moe ** 2)
    if pop:
        ss = ss / (1 + (ss-1)/pop)
    return min(ss, pop)


if __name__ == '__main__':
    r = np.arange(10)
    print wmean(r, r)
    print wstd(r, r)



import orange, Orange
import sys, math, heapq
import pdb
import numpy as np
import time

from collections import Counter

from ...util import ids_filter, block_iter

from rule import *

 
        
class BeamRefiner(object):
    def __init__(self, attrs=None, **kwargs):
        self.cost = 0.
        self.fanout = kwargs.get('fanout', 2)
        self.bdists = None
        self.ddists = None
        self.attrs = attrs or []
        self.bad_rules = set()

    def set_attributes(self, attrs=[]):
        self.attrs = attrs or []

    def add_bad_rule(self, rule):
        self.bad_rules.add( rule )

    def skip_attribute(self, attr):
        if self.attrs:
            if (attr not in self.attrs and attr.name not in self.attrs):
                return True
        return False

    def next_attributes(self, rule):
        """
        returns a generator distributions of candidate attributes to add to rule
        only returns attributes with position >= maximum attribute in current rule
        (by position in domain)
        """
        ddists = Orange.statistics.distribution.Domain(rule.examples)
        self.ddists = ddists
        bdists = self.bdists or Orange.statistics.basic.Domain(rule.examples)
        self.bdists = bdists
        attrs = rule.examples.domain
        useds = [False] * len(attrs)
        for vf in rule.filter.conditions:
            useds[vf.position] = True

        # get rule's maximum attribute (in terms of position)        
        ignore_attrs = []

        for pos, (d,b,a,u) in enumerate(zip(ddists, bdists, attrs, useds)):
            if a in ignore_attrs:
                continue            
            b = Orange.statistics.distribution.Distribution(
                a, rule.examples)
            b = Orange.statistics.distribution.Continuous(b)
            yield pos,d,b,a,u

    def construct_new_rule(self, rule, idx, ddist, bdist, attr, used, negate):
        if attr.var_type == Orange.feature.Type.Discrete:
            matches = filter(lambda c: idx == c.position,
                             rule.filter.conditions)
            if matches:
                keys = [attr.values[int(v)] for v in matches[0].values]
            else:
                keys = [k for k,v in ddist.items() if v > 0]

            if len(keys) <= 1:
                return
            

            fanout = len(keys) / 5 if len(keys) / self.fanout > 5 else self.fanout
            fanout = self.fanout
            #fanout = 60
            fanout = min(fanout, len(keys))#fanout)
            #fanout = len(keys)
            
            for keyblock in block_iter(keys, fanout):
                new_rule = rule.cloneAndAddCondition(attr, keyblock, used=used, negate=negate)
                new_rule.parent_rule = rule
                if new_rule not in self.bad_rules:
                    yield new_rule


        else:
            minv = bdist.percentile(0)
            maxv = bdist.percentile(100)
            avgv = bdist.percentile(50)
            #avgv = Orange.data.Value(bdist.variable, bdist.avg)
            #minv = Orange.data.Value(bdist.variable, bdist.min)# - 0.5
            #maxv = Orange.data.Value(bdist.variable, bdist.max)# + 0.5
            if minv == maxv:
                return

            if used:
              # This shouldn't be an issues because rule.examples
              # should filetr the distribution correctly
              minv, maxv = None, None
              for cond in rule.filter.conditions:
                if cond.position == idx:
                  minv = max(cond.min, minv) if minv else cond.min
                  maxv = min(cond.max, maxv) if maxv else cond.max

            #block = (maxv - minv) / self.fanout
            #ranges = [(minv + i*block, minv + (i+1)*block) for i in xrange(self.fanout)]
            ranges = [[minv, avgv], [avgv, maxv]]

            for minv, maxv in ranges:
                if minv == maxv: # edge case
                  continue
                new_rule = rule.cloneAndAddContCondition(
                  attr,
                  minv,
                  maxv,
                  orange.ValueFilter.Between,
                  used=used,
                  negate=negate
                )
                new_rule.parent_rule = rule
                if new_rule in self.bad_rules:
                  continue
                
                yield new_rule

        

    def __call__(self, rule, negate=False, **kwargs):
        if rule.fixed:
            return
        if not len(rule.examples):
            return

        start = time.time()        
            
        for idx, ddist, bdist, attr, used in self.next_attributes(rule):
            if self.skip_attribute(attr):
                continue

            for new_rule in self.construct_new_rule(rule, idx, ddist, bdist, attr, used, negate):
                yield attr, new_rule

        self.cost += time.time() - start



class GraphBeamRefiner(BeamRefiner):
    def __init__(self, attrs=None, graph=None, **kwargs):
        BeamRefiner.__init__(self, attrs=attrs)
        if not graph:
            raise
        self.graph = graph



    def __call__(self, rule, negate=False, extend=False, **kwargs):
        #import pdb
        #pdb.set_trace()
        if rule.fixed:
            return
        if not len(rule.examples):
            return
        

        start = time.time()        
        ddists = Orange.statistics.distribution.Domain(rule.examples)
        self.ddists = ddists
        bdists = self.bdists or Orange.statistics.basic.Domain(rule.examples)
        self.bdists = bdists
        attrs = rule.examples.domain.attributes
        useds = [False] * len(attrs)
        for vf in rule.filter.conditions:
            useds[vf.position] = True

        rule_positions = [cond.position for cond in rule.filter.conditions]
        rule_attrs = [rule.examples.domain[pos] for pos in rule_positions]

        if not len(rule_attrs):
            rule_attrs = [None]
            
        next_attrs = set(rule_attrs)        
        if extend:
            for rule_attr in rule_attrs:
                if rule_attr:
                    next_attrs.update(self.graph[rule_attr.name])
                else:
                    next_attrs.update(self.graph[rule_attr])

        for idx, (ddist, bdist, attr, used) in enumerate(zip(ddists, bdists, attrs, useds)):

            if (self.attrs and
                (attr not in self.attrs and attr.name not in self.attrs)):
                continue
            if attr not in next_attrs and attr.name not in next_attrs:
                continue

            for new_rule in self.construct_new_rule(rule, idx, ddist, bdist, attr, used, negate):
                yield attr, new_rule

# graph is a partial ordering
# graph = defaultdict(list)
# graph.update({'A': ['B', 'C'],
#               'B': ['C', 'D']})



import sys
sys.path.append('/Users/sirrice/Dropbox/code/qurkexp/qurkexp/sensitive/')
sys.path.append('..')

from datasets import data_iterator
import matplotlib.pyplot as plt
from collections import *

from scorpionsql.db import *


def load_marcua_dataset():
    for expid, (prop, answers, truth) in enumerate(data_iterator()):
        cur_wid = 0
        wids = {}
        rowid = 0
        for ans in answers:
            # normalize worker ids
            if ans[1] not in wids:
                wids[ans[1]] = cur_wid
                cur_wid += 1
            row = list(ans)
            row[1] = wids[ans[1]]
            row.append(rowid)
            rowid += 1
            row.insert(0, prop)
            row.insert(0, expid)
            row.append(truth)

            yield row


db = connect('intel')
try:
    prepare(db, 'drop table marcua')
except:
    pass

try:
    q = """create table marcua (expid int, prop varchar(128),
    est float, wid int, atime float, secs float, height int, width int, id int,
    truth float);"""
    prepare(db, q)
except:
    raise


attrs = ['expid', 'prop', 'est', 'wid', 'atime', 'secs', 'h', 'w', 'id', 'truth']
for row in load_marcua_dataset():
    try:
        q = "insert into marcua values (%s)" % (','.join(['%s'] * len(attrs)))
        prepare(db, q, tuple(row))
    except:
        import traceback
        traceback.print_exc()

db.close()



import math
import pdb
import random
import numpy as np
import orange
import Orange
import sys
sys.path.extend(['.', '..'])

from itertools import chain, product
from collections import defaultdict
from operator import sub

from ..util.prob import *
from ..util.misc import valid_number
from ..util.rangeutil import *
from ..learners.cn2sd.rule import SDRule

from bounding_box import *

nan = float('nan')
inf = float('inf')


class Cluster(object):
  _id = 0
  
  def __init__(self, bbox, error, cols, parents=None, discretes=None, **kwargs):
      """
      @param cols the set of continuous columns
      """
      self.error = error
      self.bbox = bbox and (tuple(bbox[0]), tuple(bbox[1])) or ((),())
      self.discretes = defaultdict(set)
      self.cols = cols
      self.parents = parents or []
      self.npts = kwargs.get('npts', 1)
      self.kwargs = kwargs
      self.id = Cluster._id


      self.bad_inf = None
      self.good_inf = None
      self.mean_inf = None
      self.idxs = []

      self.inf_state = None

      # data structures for merger
      # for dominance in merger

      self.inf_range = None

      # lambda c: -> influence value
      self.inf_func = None

      # range of c values this cluster is potentially
      # optimal over
      self.c_range = None
      if self.c_range is None: 
        self.c_range = [-inf, inf]

      self.states = []
      self.cards = []
      self.bad_states = []
      self.bad_cards = []
      self.good_states = []
      self.good_cards = []

      self.rule = None

      Cluster._id += 1

      if discretes:
        for k,vs in discretes.iteritems():
            self.discretes[k].update(vs)

      self.hash = None
      self._bound_hash = None

  @staticmethod
  def from_dict(thedict):
      c = Cluster([], 0., [])
      c.__dict__.update(thedict)
      c.bbox = c.bbox and (tuple(c.bbox[0]), tuple(c.bbox[1])) or ((),())
      c.cols = map(str, c.cols)
      c.discretes = dict((str(k),set([str(v) if isinstance(v,basestring) else v for v in vs])) for k,vs in c.discretes)
      return c

  def __cmp__(self, o):
    if self.error < o.error:
      return -1
    if self.error > o.error:
      return 1
    return len(str(self)) - len(str(o))

  def to_dict(self):
      d = dict(self.__dict__)
      del d['rule']
      del d['inf_func']
      if 'parents' in d:
          del d['parents']
      if 'discretes' in d:
          d['discretes'] = [(k,list(v)) for k,v in self.discretes.iteritems()]
      return d

  @property
  def root(self):
    cur = self
    while cur.parents:
      cur = cur.parents[0]
    return cur

  def __volume__(self):
      return volume(self.bbox)
  volume = property(__volume__)

  def clone(self, copy_rule=False):
      c = Cluster(
        self.bbox, self.error, self.cols,
        parents=self.parents, discretes=self.discretes, 
        **self.kwargs
      )
      c.inf_func = self.inf_func
      c.inf_state = list(self.inf_state)
      c.c_range = list(self.c_range)
      if copy_rule:
        c.rule = self.rule
      return c


  def split_on(self, c):
    """
    Split current cluster via how it intersects with c
    """
    ex_clusters = [] # clusters that don't intersect with c
    in_clusters = [] # clusters that intersect with c
    
    if not volume(intersection_box(self.bbox, c.bbox)):
        return ()

    if self.discretes:
        # they have same keys, just need to diff them
        myd = self.discretes
        od = c.discretes

        intersect_discretes = dict()
        excluded_discretes = dict()
        for key in self.discretes:
            myvals = set(myd[key])
            ovals = set(od[key])
            intersect_discretes[key] = myvals.intersection(ovals)
            excluded_discretes[key] = myvals.difference(ovals)

        if not sum(map(len, intersect_discretes.values())):
            return ()

        if not sum(map(len, excluded_discretes.values())):
            return ([self.clone()], [])


        excluded_cluster = self.clone() 
        excluded_cluster.discretes = excluded_discretes
        ex_clusters.append(excluded_cluster)
    else:
        intersect_discretes = dict(self.discretes)

    if box_contained(self.bbox, c.bbox):
        intersect_cluster = self.clone()
        intersect_cluster.discretes = intersect_discretes
        return ([intersect_cluster], ex_clusters)

    # how the painful part.. splitting the contiuous domain
    inner_mins = map(max, zip(self.bbox[0], c.bbox[0]))
    inner_maxs = map(min, zip(self.bbox[1], c.bbox[1]))
    all_ranges = [[] for i in inner_mins]
    for idx, (bmin, bmax, imin, imax) in enumerate(zip(self.bbox[0], self.bbox[1], inner_mins, inner_maxs)):
        vals = sorted(set([bmin, bmax, imin, imax]))
        for i in xrange(len(vals)-1):
            all_ranges[idx].append((vals[i], vals[i+1]))

    all_boxes = product(*all_ranges)
    groups = defaultdict(list)
    for box in all_boxes:
        new_cluster = self.clone()
        new_cluster.bbox = tuple(zip(*box))
        new_cluster.discretes = dict(intersect_discretes)
        groups[box_contained(new_cluster.bbox, c.bbox)].append(new_cluster)

    ex_clusters.extend(groups.get(False, []))
    in_clusters = groups.get(True, [])
    for ec in ex_clusters:
        ec.parents = (self, c)
    return (list(in_clusters), list(ex_clusters))


  @staticmethod
  def merge(c1, c2, intersecting_clusters=None, min_volume=0):
    """
    Computes the cluster after merging c1 and c2

    If intersecting clusters is not empty, computes a
    weighed estimated error based on the intersecting clusters
    """
    if c1.cols != c2.cols:
      raise RuntimeError("columns do not match!  %s vs %s" % (str(c1.cols), str(c2.cols)))
    
    # Compute the new cluster
    newbbox = bounding_box(c1.bbox, c2.bbox)
    discretes = defaultdict(set)
    for k in chain(c1.discretes.keys(), c2.discretes.keys()):
        discretes[k].update(c1.discretes.get(k, ()))
        discretes[k].update(c2.discretes.get(k, ()))


    if not intersecting_clusters:
      newerror = float('-inf')
    else:
      # compute new error
      weights = []
      errors = []
      for intersection in intersecting_clusters:
          bbox = intersection_box(intersection.bbox, newbbox)
          weights.append(volume(bbox) or min_volume)
          errors.append(intersection.error)

      total_weight = sum(weights) or 1.
      if not total_weight:
          return None

      newerror = sum(e*w for e,w in zip(errors, weights)) / total_weight

    return Cluster(newbbox, newerror, c1.cols, parents=(c1, c2), discretes=discretes)


  @staticmethod
  def from_rule(rule, cols, error=None, parents=None):
    """
    @param cols list of attribute names
    """
    domain = rule.data.domain
    name_to_bounds = {}
    discretes = {}
    
    for c in rule.filter.conditions:
      attr = domain[c.position]
      name = attr.name
      if attr.varType == orange.VarTypes.Discrete:
        if name in cols:
          try:
            discretes[name] = [int(v) for v in c.values]
          except Exception as e:
            print e
            print c.values
      else:
        name_to_bounds[name] = (c.min, c.max)

    mins, maxs = [], []
    cont_cols = []
    for col in cols:
      if col in name_to_bounds:
        mins.append(name_to_bounds[col][0])
        maxs.append(name_to_bounds[col][1])
        cont_cols.append(col)

    bbox = (tuple(mins), tuple(maxs))
    error = error or rule.quality
    cluster = Cluster(bbox, error, cont_cols, parents=parents, discretes=discretes, npts=len(rule.examples))
    cluster.rule = rule
    cluster.c_range = rule.c_range
    cluster.inf_state = rule.inf_state
    return cluster

  def adjacent(self, o, thresh=0.7):
    """@return True if they overlap on one or more continuous attribute, 
        and their discrete attributes intersect"""
    d_intersects =  self.discretes_intersect(o)
    if not d_intersects:
        return False

    # needs to "sufficiently" intersect on N-1 dimensions and
    # at least touch on one dimension
    N = len(self.bbox[0])
    n_intersect = 0
    n_close = 0
    dists = []
    for myb, ob in  zip(zip(*self.bbox), zip(*o.bbox)):
        _min, _max = max(myb[0], ob[0]), min(myb[1], ob[1])
        myrange = myb[1] - myb[0]
        dist = _max - _min
        dists.append(dist)
        if dist > myrange * 0.1:
            n_intersect += 1
        elif dist > -myrange * 0.01:
            n_close += 1

    return n_intersect + n_close == N and n_close <= 1

  def inf_dominates(self, o, alpha=0, c_range=None):
    """
    Sample c values along both influence curves and ensure that
    this cluster's influence values are 
    """
    if c_range is None:
      c_range = r_union(self.c_range, o.c_range)
    xs = ((np.arange(20)/20.) * r_vol(c_range)) + c_range[0]
    mine = self.inf_func(xs)
    theirs = o.inf_func(xs) * (1.0 + alpha)
    return (mine >= theirs).all()

  def contains(self, o, epsilon=0.):
    if not box_contained(o.bbox, self.bbox, epsilon=epsilon):
      return False
    return self.discretes_contains(o)

  def same(self, o, epsilon=0.):
    for key in o.discretes.keys():
      if key in self.discretes:
        diff = set(o.discretes[key]).intersection(self.discretes[key])
        if len(diff) != len(o.discretes[key]) or len(diff) != len(self.discretes[key]):
          return False

    return box_same(o.bbox, self.bbox, epsilon=epsilon)

  def discrete_differences(self, o, epsilon=0.):
    """
    Returns None if o.bbox is different than self.bbox
            otherwise, dictionary of:
              attr -> set(values self doesn't have)
    """
    if not box_same(o.bbox, self.bbox, epsilon=epsilon):
      return None

    ret = {}
    for key, ovals in o.discretes.iteritems():
      vals = ovals.difference(self.discretes.get(key, ()))
      if vals:
        ret[key] = vals
    return ret


  def discretes_contains(self, o):
      for key in o.discretes.keys():
          if key in self.discretes:
              diff = set(o.discretes[key]).difference(self.discretes[key])
              if len(diff):
                  return False
      return True

  def discretes_same(self, o):
      mykeys = set(self.discretes.keys())
      okeys = set(o.discretes.keys())
      if mykeys != okeys:
          return False
      for key in mykeys:
          if set(self.discretes[key]) != set(o.discretes[key]):
              return False
      return True

  def discretes_overlap(self, o):
      myd = self.discretes
      od = o.discretes
      keys = set(myd.keys()).intersection(set(od.keys()))
      
      for key in keys:
          if len(od[key].intersection(myd[key])) == 0:
              return False
      return True

  def discretes_intersect(self, o):
      myd = self.discretes
      od = o.discretes
      keys = set(myd.keys()).intersection(set(od.keys()))
      
      for key in keys:
          if len(od[key].intersection(myd[key])) < min(len(od[key]), len(myd[key])):
              return False
      return True

  def discretes_distance(self, o):
      myd = self.discretes
      od = o.discretes
      keys = set(myd.keys()).union(set(od.keys()))
      
      diff = 0
      for key in keys:
          mvals = myd.get(key, set())
          ovals = od.get(key, set())
          diff += len(mvals) + len(ovals) - len(mvals.intersection(ovals))
      return diff



  def to_rule(self, table, cont_dists=None, disc_dists=None):
    """
    @param cols list of attribute names
    """
    if not self.rule:
      domain = table.domain
      attrnames = [attr.name for attr in domain]
      cont_dists = cont_dists or dict(zip(attrnames, Orange.statistics.basic.Domain(table)))
      disc_dists = disc_dists or dict(zip(attrnames, Orange.statistics.distribution.Domain(table)))
      conds = []

      for col, bound in zip(self.cols, zip(*self.bbox)):
        attr = domain[col]
        pos = domain.index(attr)
        table_bound = cont_dists[attr.name]
        minv, maxv = r_intersect(bound, [table_bound.min, table_bound.max])
        if maxv - minv > 0.99 * (table_bound.max-table_bound.min):
          continue
        
        conds.append(
          orange.ValueFilter_continuous(
              position=pos,
              max=bound[1],
              min=bound[0]
          )
        )

      for disc_name, vidxs in self.discretes.iteritems():
        attr = domain[disc_name]
        disc_pos = domain.index(attr)
        vals = [orange.Value(attr, attr.values[int(vidx)]) for vidx in vidxs if int(vidx) < len(attr.values)]

        if not vals or len(vals) == len(disc_dists[attr.name]):
          continue
        
        conds.append(
          orange.ValueFilter_discrete(
              position = disc_pos,
              values=vals
          )
        )

      rule = SDRule(table, None, conditions=conds)
      self.rule = rule

    rule = self.rule
    rule.quality = rule.score = self.error
    rule.inf_state = self.inf_state
    rule.c_range = self.c_range
    return rule

  def project(self, cols):
      """
      @param cols all of the columns (continuous and discrete) to keep
      """
      newbbox = [[], []]
      newcols = []
      for col, minv, maxv in zip(self.cols, self.bbox[0], self.bbox[1]):
          if col in cols:
              newbbox[0].append(minv)
              newbbox[1].append(maxv)
              newcols.append(col)

      newdiscretes = dict(filter(lambda (k,v): k in cols, self.discretes.iteritems()))
      return Cluster(newbbox, self.error, newcols, parents=self.parents, discretes=newdiscretes)



  def __ancestors__(self):
      if not self.parents:
          return (self,)
      ret = []
      for p in self.parents:
          ret.extend(p.ancestors)
      return ret
  ancestors = property(__ancestors__)
  
  def __str__(self):
      cr = "%.4f-%.4f" % tuple(self.c_range)
      return '%.4f\t%s\t%s' % (self.error, cr, str(self.rule))
      s = '\t'.join(['%s:(%.4f, %.4f)' % (col, bound[0], bound[1]) 
                  for col, bound in zip(self.cols, zip(*self.bbox))])
      d = '\t'.join(["%s:%s (%d els)" % (k, str(list(v)[:3]), len(v)) 
        for k,v in self.discretes.iteritems()])
      return '%.4f\t%s\t%s\t%s' % (self.error, cr, s, d)
      fmt = '\t'.join(['%.4f'] * len(self.bbox[0]))
      args = (self.error, fmt % self.bbox[0], fmt % self.bbox[1], str(self.discretes))
      return '%.6f\t%s\t%s\t%s' % args

  def __hash__(self):
    if self.hash is None:
      self.hash = self.bound_hash
    return self.hash


  @property
  def bound_hash(self):
    if self._bound_hash is None:
      discretes = []
      for dkey in sorted(self.discretes.keys()):
        discretes.append(dkey)
        discretes.append(":::")
        discretes.extend(map(hash, sorted(self.discretes[dkey]))) 
      self._bound_hash = hash((self.bbox, str(discretes)))
    return self._bound_hash

  def __eq__(self, o):
      return hash(self) == hash(o)


def filter_bad_clusters(clusters):
  f = lambda c: c and valid_number(c.error)
  return filter(f, clusters)

def compute_clusters_threshold(clusters, nstds=1.):
    if not clusters:
        return 0.
    errors = [c.error for c in filter_bad_clusters(clusters)]
    return np.percentile(errors, 70)
    npts = [c.npts for c in clusters]
    #npts = [1] * len(clusters)
    mean, std = wmean(errors, npts), wstd(errors, npts)        
    thresh = max(0, min(max(errors), mean + nstds * std))
    thresh = min(max(errors), mean + nstds * std)
    return thresh


def filter_top_clusters(clusters, nstds=1.):
    """
    compute mean and std of error of clusters.  return clusters
    nstds above mean
    """
    if len(clusters) <= 1:
        return clusters
    thresh = compute_clusters_threshold(clusters, nstds)
    f = lambda c: c.error >= thresh
    return filter(f, clusters)

    errors = [c.error for c in clusters]
    minv, maxv, mean, std = min(errors), max(errors), np.mean(errors), np.std(errors)
    thresh = min(maxv, mean + nstds * std)
    thresh = 0.6 * (maxv - minv) + minv
    f = lambda c: c.error >= thresh
    return filter(f, clusters)


def normalize_cluster_errors(clusters):
    if not clusters:
        return clusters
    errors = np.array([c.error for c in clusters])
    mean, std = np.mean(errors), np.std(errors)
    if std == 0: std = 1
    for c in clusters:
      c.error = (c.error - mean) / std
    return clusters
    mine, maxe = min(errors), max(errors)
    div = 1. if maxe == mine else (maxe-mine)

    for c in clusters:
        c.error = (c.error - mine) / div
    return clusters


def clusters_to_rules(clusters, table):
    import Orange
    attrnames = [attr.name for attr in table.domain]
    cont_dists = dict(zip(attrnames, Orange.statistics.basic.Domain(table)))
    disc_dists = dict(zip(attrnames, Orange.statistics.distribution.Domain(table)))
    args = {'cont_dists' : cont_dists,
            'disc_dists' : disc_dists}

    rules = [c.to_rule(table, **args) for c in clusters]

    return rules



import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

from itertools import chain


from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *
from ..settings import *

from merger.merger import Merger
from merger.reexec import ReexecMerger
from basic import Basic
from sampler import Sampler

inf = 1e10000000





class Splits(object):
    def __init__(self, influence):
        self.splits = []
        self.influence = influence

    def __stop__(self):
        return len(self.splits) == 0
    stop = property(__stop__)

    def add_split(self, s):
        self.splits.append(s)

class Split(object):
    def __init__(self, attr, *args):
        self.attr = attr
        self.stop = attr is None
        if self.stop:
            self.influence = args[0]
        else:
            self.rules = args[0]
            self.score = args[1]



class BDTPartitioner(Basic):
    def set_params(self, **kwargs):
        Basic.set_params(self, **kwargs)

        self.p = kwargs.get('p', 0.6)
        self.tau = kwargs.get('tau', [0.001, 0.05])
        self.epsilon = kwargs.get('epsilon', 0.005)
        self.min_pts = 5
        self.samp_rate = 1.
        self.SCORE_ID = kwargs['SCORE_ID']
        self.inf_bounds = [inf, -inf]

        self.sampler = Sampler(self.SCORE_ID)


    def setup_table(self, table):
        self.table = table
        self.err_func.setup(table)

   
    def __call__(self, table, **kwargs):
        self.setup_table(table)

        base_rule = SDRule(table, None)
        node = self.grow(base_rule, self.samp_rate)

        return node.leaves


    def grow(self, rule, samp_rate):
        data = rule.examples
        sample = self.sample(data, samp_rate)
        node = Node(rule)

        if self.should_stop(sample):
            node.set_score(self.estimate_inf(sample))
            return node

        scores = []
        for attr, new_rules in self.child_rules(rule):
            score = self.get_score(new_rules, sample)
            scores.append(new_rules, score)
        
        if not scores:
            node.set_score(self.estimate_inf(sample))
            return node

        new_rules, score = min(scores, key=lambda pair: p[1])
        if len(new_rules) == 1:
            return self.grow(new_rules, samp_rate)


        new_sample_rates = self.update_sample_rates(new_rules, samp_rate)
        for new_rule, new_samp_rate in zip(new_rules, new_sample_rates):
            child = self.grow(new_rule, new_samp_rate)
            if child.n:
                node.add_child(child)

        return node

    def estimate_inf(self, sample):
        return np.mean(map(self.influence, sample))
        
    def get_score(self, new_rules, sample):
        new_samples = map(lambda r: r.filter_table(sample), new_rules)
        score = sum(map(self.compute_score, new_samples))
        return score

    def should_stop(self, data):
        if len(data) < self.min_pts:
            return True

        infmax = max(map(self.influence, data))
        thresh = self.compute_threshold(infmax, self.inf_bounds[0], self.inf_bounds[1])
        score = self.compute_score(data)
        return score < thresh

    def compute_score(self, data):
        return np.std(map(self.influence, data))


    def child_rules(self, rule, attrs=None):
        attrs = attrs or self.cols
        next_rules = defaultdict(list)
        refiner = BeamRefiner(attrs=attrs, fanout=2)
        for attr, new_rule in refiner(rule):
            next_rules[attr].append(new_rule)
        return next_rules.items()


    def sample(self, data, samp_rate):
        return self.sampler(data, samp_rate)


    def update_sample_rates(self, rules, samp_rate):
        influences, counts = [], []
        for rule in rules:
            influence = sum(map(self.influence, rule.examples))
            influences.append(influence)
            counts.append(float(len(rule.examples)))

        total_inf = sum(influences)
        total_count = sum(counts)
        samp_rates = []
        for influence, count in zip(influences, counts):
            nsr = samp_rate * (influence/total_inf) * (count/total_count)
            samp_rates.append(nsr)

        return samp_rates



    def compute_threshold(self, infmax, infl, infu):
        tau, p = self.tau, self.p
        s = (tau[0] - tau[1]) / ((1-p)*infu - p * infl)
        w = tau[0] + s*(infu - infmax)
        w = min(tau[1], w)
        return w * (infu - infl)       

    def influence(self, row):
        if row[self.SCORE_ID].value == -inf:
            influence = self.err_func((row,))
            row[self.SCORE_ID] = influence
            self.inf_bounds[0] = min(influence, self.inf_bounds[0])
            self.inf_bounds[1] = max(influence, self.inf_bounds[1])
        return row[self.SCORE_ID].value





#
# For sigmod SYNTH Tests
# 1) compute synth for 2/3/4-d with medium and high outlier boxes
# 2) compute the two c values and store it
# 3) compute and store the boxes
# 4) execute suite of tests for each dataset
#
import pdb
import sys
import math
import json
import random
import numpy as np

from sqlalchemy import create_engine
from scorpion.util import get_logger
from scorpion.misc.gensinglecluster import gen_points, in_box


from common import *
from setup_experiments import init_db

random.seed(0)

format_list = lambda fmt, l: [fmt % (isinstance(p, list) and tuple(p) or p) for p in l]


def generate_datasets(db, ndim, uo=30, npts=2000):
    random.seed(0)
    """
    generates the 2/3/4-d datasets and stores them in db: sigmod
    """
    mid_box, high_box, schema, generator = gen_points(npts, ndim, ndim, 0.25, 10, 10, uo, 10)
    pts = [pt for pt in generator]
    c_mid, c_high = compute_c_values(pts, mid_box, high_box)

    # store boundaries
    tablename = 'data_%d_%d' % (ndim, int(uo))
    id = add_config(db, tablename, 2000, ndim, ndim, 0.25, 10, 10, uo, 10, mid_box, high_box, c_high)

    # save the damn thing
    save_pts(db, tablename, pts)
    

def setup(db):
    try:
        with db.begin() as conn:
            conn.execute("""
            create table synth (
                id serial,
                tablename varchar(128),
                npts int,
                ndim int,
                kdim int,
                volperc float,
                uh float,
                sh float,
                uo float,
                so float,
                mid text,  -- json encoded boundary for mid level
                high text,  -- json encoded
                c float -- threshold for picking  high over mid
            )
            """)
    except:
        import traceback
        traceback.print_exc()
        pass

def save_pts(db, tablename, pts):
    ndims = len(pts[0]) - 2
    with db.begin() as conn:
        q = """
        create table %s (
        id serial,
        %s,
        g float,
        v float)""" % (tablename, ','.join(format_list('a%d float', xrange(ndims))))
        conn.execute(q)

        pts_str = []
        for pt in pts:
            s = '(%s)' % ','.join(['%f']*(ndims+2)) % tuple(pt)
            pts_str.append(s)


        bigq = """insert into %s (%s) values
           %s""" % (tablename, ','.join(format_list('a%d', xrange(ndims))+['g','v']), ','.join(pts_str))
        conn.execute(bigq)


def add_config(db, *args):
    args = list(args)
    with db.begin() as conn:
        q = """
        insert into synth (tablename, npts, ndim, kdim, volperc, uh, sh, uo, so, mid, high, c)
        values (%s) returning id""" % ((',%s' * 12)[1:])
        args[-3] = json.dumps(args[-3])
        args[-2] = json.dumps(args[-2])
        id = conn.execute(q, *args ).fetchone()[0]
        return id

def all_configs(db):
    with db.begin() as conn:
        q = """ select * from synth"""
        for config in conn.execute(q).fetchall():
            config = list(config)
            config[-3] = json.loads(config[-3])
            config[-2] = json.loads(config[-2])
            yield config



def get_config(db, tablename):
    with db.begin() as conn:
        q = """ select * from synth where tablename = %s"""
        config = conn.execute(q, tablename).fetchone()
        config = list(config)
        config[-3] = json.loads(config[-3])
        config[-2] = json.loads(config[-2])
        return config

def get_pts(db, tablename):
    with db.begin() as conn:
        q = """ select * from %s""" % tablename
        return [pt[1:] for pt in conn.execute(q).fetchall()]



def compute_c_values(pts, mid_bounds, high_bounds, f=np.mean):
  """
  @deprecated don't use
  """
  return 0, 1

  orig_f = f
  f = lambda pts: orig_f([pt[-1] for pt in pts])
  pts = set(map(tuple,pts))
  all_vs, mid_vs, high_vs = set(), set(), set()
  n_mid, n_high = 0, 0
  for pt in pts:
      if pt[-2] != 8:
          continue
      all_vs.add(pt)
      if in_box(pt[:-2], mid_bounds):
          mid_vs.add(pt)
      if in_box(pt[:-2], high_bounds):
          high_vs.add(pt)


  orig = f(all_vs)
  nomid = f(all_vs.difference(mid_vs))
  nohigh = f(all_vs.difference(high_vs))

  dm = orig - nomid
  dh = orig - nohigh

  cm = len(mid_vs)
  ch = len(high_vs)
  
  highc = math.log(dh / dm) / math.log(1.*ch/cm)
  
  stats = []
  for c in xrange(0, 50, 5):
      c = c / 100.
      minfs, hinfs = [], []
      for n in xrange(5000):
          mid_samp = random.sample(mid_vs.difference(high_vs), random.randint(1, 100))
          high_samp = random.sample(high_vs, random.randint(1, 100))
          mix_samp = mid_samp + high_samp
          dm = orig - f(all_vs.difference(mix_samp)) 
          dh = orig - f(all_vs.difference(high_samp))
          ch, cm = float(len(high_samp)), float(len(mix_samp))
          if dh <= 0 or dm <= 0 or dh == dm or ch == cm:
              continue

          minfs.append(dm / cm**c)
          hinfs.append(dh / ch**c)

      mmean, mstd = np.mean(minfs), np.std(minfs)
      hmean, hstd = np.mean(hinfs), np.std(hinfs)
      stats.append((c, mmean, mstd, hmean, hstd))
      print ('\t%.4f'*5) % stats[-1]
  return highc - .25*np.std(highcs), highc + .25*np.std(highcs)

  return highc

def save_result(db, total_cost, costs, stats, rule, ids, dataset, notes, kwargs):
  acc, p, r, f1 = stats
  ids_str = ','.join(map(str, ids))
  isbest = rule.isbest
  vals = [0, dataset, notes]
  pkeys = ['klassname', 'cols', 'epsilon', 'c', 'lambda', 'boundtype']
  vals.extend([kwargs.get(key, None) for key in pkeys])
  vals.extend((total_cost, acc, p, r, f1, rule.quality, isbest, str(rule), ids_str))
  stat = vals


  with db.begin() as conn:
      q = """insert into stats(expid, dataset, notes, klass, cols, epsilon, c,     
                                lambda, boundtype, cost, acc, prec, recall, f1, score, 
                                isbest, rule, ids) values(%s) returning id""" % (','.join(['%s']*len(stat)))
      sid = conn.execute(q, *stat).fetchone()[0]

      q = """insert into costs(sid, name, cost) values(%s,%s,%s)"""
      for name, cost in costs.items():
          if isinstance(cost, list):
              cost = cost[0]
          conn.execute(q, sid, name, cost)
      return sid



def compute_stats_from_bounds(sigmoddb, learner, dataset, bounds, table_size, ids):
    try:
        truth = set(get_ids_in_bounds(sigmoddb, dataset, bounds))
        return compute_stats(ids, truth,  table_size)
    except:
        print "defaulting"
        return learner.compute_stats(ids)

def compute_all_stats_from_bounds(sigmoddb, learner, dataset, bounds, table_size, all_ids):
    try:
        truth = set(get_ids_in_bounds(sigmoddb, dataset, bounds))
        return [compute_stats(ids, truth,  table_size) for ids in all_ids]
    except:
        print "defaulting"
        return map(learner.compute_stats,all_ids)




def run(sigmoddb, statsdb, tablename, all_bounds, **kwargs):

    try:
        os.system('rm *.cache')
    except:
        pass


    params = {
            'epsilon' : 0.0005,
            'tau' : [0.1, 0.5],
            'p' : 0.7,
            'l' : 0.5,
            'min_improvement' : 0.01,
            'c' : 0.0,
            'max_wait' : 20 * 60,
            'klass' : BDT,
            'granularity' : 20
            }

    params.update(kwargs)
    klassname = params['klass'].__name__
    params['klassname'] = klassname
    params['dataset'] = tablename
    params['use_mtuples'] = params.get('use_mtuples', True) and klassname == 'BDT'
    cs = params['cs']




    if klassname == 'Naive':
        print "running experiment"
        costs, rules, all_ids, table_size, learner = run_experiment(tablename, **params)
        print "computing stats and saving"
        for bounds, boundtype in all_bounds:
            params['boundtype'] = boundtype


            for c, rules in learner.bests_per_c.items():
                if not rules: continue
                params['c'] = c
                rule = max(rules, key=lambda r: r.quality)
                ids = get_row_ids(rule, learner.full_table)
                stats = compute_stats_from_bounds(sigmoddb, learner, tablename, bounds, table_size, ids)
                save_result(statsdb, costs['cost_total'], costs, stats, rule, ids, tablename, str(bounds), params)

            for c, checkpoints in learner.checkpoints_per_c.items():
                params['c'] = c
                for timein, rule in checkpoints:
                    chk_ids = get_row_ids(rule, learner.full_table)
                    stats = compute_stats_from_bounds(sigmoddb, learner,  tablename, bounds, table_size, chk_ids)
                    save_result(statsdb, timein, {}, stats, rule, chk_ids, tablename, str(bounds), params)
        print "done"
    else:
        for c in cs:
            params['c'] = c
            print "running experiment"
            costs, rules, all_ids, table_size, learner = run_experiment(tablename, **params)
            print "computing stats and saving"

            for bounds, boundtype in all_bounds:
                params['boundtype'] = boundtype
#                truth = set(get_ids_in_bounds(sigmoddb, tablename, bounds))
#                all_stats = [compute_stats(ids, truth,  table_size) for ids in all_ids]
                all_stats = compute_all_stats_from_bounds(sigmoddb, learner, tablename, bounds, table_size, all_ids)
                for stats, rule, ids in zip(all_stats, rules, all_ids):
                    save_result(statsdb, costs['cost_total'], costs, stats, rule, ids, tablename, str(bounds), params)
            print "done"


def run_tests(sigmoddb, statsdb, **params):
    """
    These recreate the synthetic tests that use 2k pts per group, and 
    vary dimensionality and easy vs hard.  These are all graphs
    section 8.3 except figures 15 and 16

    defaults: 
     epsilon: 0.001
     tau: [0.1, 0.5]
     p: 0.7
     lambda: 0.5
     min_improvement: 0.01
    """

    for ndim in [2,3,4]:
        for uo in [30, 80]:

            tablename = "data_%d_%d" % (ndim, uo)
            config = get_config(sigmoddb, tablename)
            mid_bounds = config[-3]
            high_bounds = config[-2]
            all_bounds = [(mid_bounds, 'mid'), (high_bounds, 'high')]
            pts = get_pts(sigmoddb, tablename)
            cs = reversed([0., 0.05, 0.075, 0.09, 0.1, 0.15, 0.2, 0.5])
            #cs = [0.5, 0.2]

            for klass in [Naive, BDT, MR]:
                run(sigmoddb, statsdb, tablename, all_bounds, cs=cs, klass=klass, **params)

statsdb = create_engine('postgresql://localhost/dbwipes')
sigmoddb = create_engine('postgresql://localhost/sigmod')
init_db(statsdb)



for dataset in ['0', '11']:
    for klass in [MR,BDT,NDT]:
        run(sigmoddb, statsdb, dataset, [(None, 'none')], 
            cs=[0,0.05,0.075,0.09,0.1,0.15,0.2,0.5], 
            klass=klass, 
            granularity=15, 
            use_mtuples=True, 
            use_cache=False
        )
exit()

#tablename = 'data_2_30'
#config = get_config(sigmoddb, tablename)
#pts = get_pts(sigmoddb, tablename)
#run_tests(sigmoddb, statsdb, max_wait=12*60, granularity=15, use_mtuples=True, use_cache=False)

if False:
    setup(sigmoddb)
    for uo in [30, 80]:#40, 50, 60, 70, 80, 90, 100]:
        generate_datasets(sigmoddb, 2, uo=uo)
        generate_datasets(sigmoddb, 3, uo=uo)
        generate_datasets(sigmoddb, 4, uo=uo)



import pdb
import sys
import random
import time
import sys
import matplotlib
import numpy as np
sys.path.extend( ['.', '..'])

from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.patches import Rectangle as Rect
from matplotlib import cm 
import matplotlib.pyplot as plt
from collections import defaultdict
from sqlalchemy import create_engine

from scorpionsql.db import *
from scorpionsql.aggerror import *

from arch import *
from gentestdata import *
from util import reconcile_tables
from sigmod import *
from common import *

matplotlib.use("Agg")



def print_clusters(pp, clusters, tuples=[], title=''):
    fig = plt.figure(figsize=(8, 8))
    sub = fig.add_subplot(111)
    clusters.sort(key=lambda c: c.error)

    for cluster in clusters:
      x, y = tuple(map(list, zip(*cluster.bbox)))[:2]
      x[0] = max(0, x[0])
      x[1] = min(100, x[1])
      y[0] = max(0, y[0])
      y[1] = min(100, y[1])
      c = cm.jet(cluster.error)
      r = Rect((x[0], y[0]), x[1]-x[0], y[1]-y[0], alpha=max(0.1,cluster.error), ec=c, fill=False, lw=1.5)
      sub.add_patch(r)

    if tuples:
      cols = zip(*tuples)
      xs, ys, cs = cols[0], cols[1], cols[-1]
      sub.scatter(ys, xs, c=cs, alpha=0.5, lw=0)


    sub.set_ylim(-5, 105)
    sub.set_xlim(-5, 105)
    sub.set_title(title)
    plt.savefig(pp, format='pdf')



def run(pp, datasetidx, bounds, **params):
  test_datas = get_test_data(datasetnames[datasetidx])
  tablename = test_datas[-1]
  dbname = test_datas[0]
  db = create_engine('postgresql://localhost/%s' % dbname)

  costs, rules, all_ids, table_size, learner = run_experiment(datasetidx, **params)
  cost = costs['cost_total']


  try:
    truth = set(get_ids_in_bounds(db, tablename, bounds))
    all_stats = [compute_stats(ids, truth,  table_size) for ids in all_ids]
  except Exception as e:
    all_stats = map(learner.compute_stats, all_ids)

  #  print "\n".join(map(str, learner.costs.items()))
  for stats, rule, ids in zip(all_stats, rules, all_ids)[:1]:
    print "stats:%s,c(%.3f),cost(%.2f),%.6f,%.6f,%.6f,%.6f" % tuple([tablename,params['c'],cost]+list(stats))
    print 'stats:%s'% str(sdrule_to_clauses(rule.simplify())[0])

  try:
    clusters = normalize_cluster_errors([c.clone() for c in learner.final_clusters])
    best_clusters = sorted(clusters, key=lambda c: c.error, reverse=True)[:1]


    tuples = get_tuples_in_bounds(db, tablename, [(0,100), (0,100)], 'g = 7')
    cols = zip(*tuples)

    for idx in xrange(len(cols)-4):
      tuples = zip(cols[idx], cols[idx+1], np.array(cols[-2])/100.)
      print_clusters(pp, clusters,  tuples=tuples,title="merged clusters %.4f dim %d" % (params['c'], idx))
      print_clusters(pp, best_clusters[:10],  tuples=tuples, title="best clusters %.4f dim %d" % (params['c'], idx))
  except Exception as e:
    print e
    pdb.set_trace()
    pass




       

if __name__ == '__main__':
    

  np.seterr(all='raise')
  nbadresults = 10
  idxs = sys.argv[1:] or [0,1]
  bounds = [[42.2210925762524, 92.2210925762524], [37.89772014701512, 87.89772014701512]] 
  #bounds = [[52.73538209702353, 77.73538209702353], [44.3706389043392, 69.3706389043392]]
  bounds = [[24.73254341295866, 95.44322153161342], [22.19997047910137, 92.91064859775612], [12.31825640510083, 83.02893452375558], [7.583496039802494, 78.29417415845725]]
  cs = reversed([0., 0.05, 0.075, 0.09, 0.1, 0.12, 0.15, 0.2, 0.3, 0.5, 0.75, 1.])
  cs = [0.25, 0.15, 0.125, 0.1, 0.05]
  cs = [0.25, 0.1]
  for datasetidx in idxs:
    pp = PdfPages('figs/topdown_all_%s.pdf' % str(datasetidx))
    for c in cs:
      run(pp, datasetidx, bounds,
        klass=BDT, 
        nbadresults = nbadresults,
        epsilon=0.0015,
        tau=[0.1, 0.5],
        p = 0.7,
        l=.5,
        min_pts = 10,
        min_improvement=.01,
        granularity=15,
        max_wait=1,#5*60,#None,
        naive=False,
        use_mtuples=False,
        use_cache=True,
        cs=cs,
        c=c
      )

    pp.close()

#
#  cluster = None
#  while True:
#      print "set cluster to a value"
#      pdb.set_trace()
#      if not cluster:
#          break
#      pp = PdfPages('figs/topdown_%s.pdf' % outname)
#      topdown.merger.adj_matrix.insert(cluster)
#      neighbors = topdown.merger.adj_matrix.neighbors(cluster)
#      for n in neighbors:
#          n.error = 0.5
#      cluster.error = 1
#      print_clusters(pp, list(neighbors) + [cluster], title='foo')
#
#      pp.close()
#
#





import json
import math
import pdb
import random
import numpy as np
import sys
import time
sys.path.extend(['.', '..'])

from itertools import chain, repeat, izip
from collections import defaultdict
from operator import mul, and_, or_
from scipy.optimize import fsolve
from matplotlib import pyplot as plt

from ..util import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *

from frontier import *
from merger import Merger

_logger = get_logger()



class BaseRangeMerger(Merger):
  """
  Merges clusters

  transparently scales
  - cluster bounding boxes by table bounds
  - errors by min/max error
  """


  def __init__(self, *args, **kwargs):
    Merger.__init__(self, *args, **kwargs)

    self.learner_hash = kwargs.get('learner_hash', '')
    self.c_range = kwargs.get('c_range', [0.01, 0.7])
    self.get_frontier = Frontier(self.c_range, 0.)
    self.CACHENAME = './dbwipes.rangemerger.cache'
    self.i = 0


    #
    # per execution state
    #

    # dim -> list of value subsets that were not on frontier
    # e.g., subregion -> [ (SR1, SR2), (SR3), ... ]
    self.rejected_disc_vals = defaultdict(list)

    # (dim, direction) -> range it has expanded along
    self.rejected_cont_vals = defaultdict(set)

  def __hash__(self):
    return hash((self.learner_hash, tuple(self.c_range)))

  def setup_stats(self, clusters):
    """
    computes error bounds and the minimum volume of a 0-volume cluster

    adds data structures to cluster object
    """
    Merger.setup_stats(self, clusters)

    for c in clusters:
      c.inf_func = self.learner.create_inf_func(c)
      c.c_range = list(self.c_range)
      c.inf_range = [c.inf_func(c.c_range[0]), c.inf_func(c.c_range[1])]

  def print_clusters(self, clusters):
    print "\n".join(map(str, clusters))

  def get_variety_frontier(self, clusters):
    seen = set()
    ret = []
    cluster_bcs = [list(c.inf_state[1]) for c in clusters]

    while True:
      for c in clusters:
        c.c_range = list(self.c_range)

      frontier,_ = self.get_frontier(clusters)
      frontier.difference_update(seen)
      print "variety"
      self.print_clusters(frontier)
      if not frontier: break
      ret.append(frontier)
      seen.update(frontier)

      for c in clusters:
        for f in frontier:
          for idx in xrange(len(c.inf_state[1])):
            c.inf_state[1][idx] = max(0, c.inf_state[1][idx] - 3*f.inf_state[1][idx])

    for idx, c in enumerate(clusters):
      c.inf_state = list(c.inf_state)
      c.inf_state[1] = cluster_bcs[idx]
    return ret, set(clusters).difference(seen)


  @instrument
  def __call__(self, clusters, **kwargs):
    if not clusters:
        return list(clusters)

    _logger.debug("merging %d clusters", len(clusters))
    _logger.debug("DEBUG = %s", self.DEBUG)

    self.set_params(**kwargs)
    self.setup_stats(clusters)
    self.rejected_disc_vals = defaultdict(list)
    self.rejected_cont_vals = defaultdict(set)


    self.learner.update_status("expanding frontier: indexing partitions")
    self.adj_graph = self.make_adjacency(clusters, self.partitions_complete)

    if self.DEBUG:
      self.renderer = InfRenderer('/tmp/infs.pdf', c_range=self.c_range)
      self.renderer.plot_inf_curves(clusters)
      self.renderer.set_title("inf 0")


    self.learner.update_status("expanding frontier, iter 0")
    frontiers, removed_clusters = self.get_variety_frontier(clusters)
    _logger.debug("%d clusters in frontier", sum(map(len, frontiers)))

    clusters.sort(key=lambda c: c.inf_func(0.1), reverse=True)
    self.print_clusters(chain(*frontiers))
    print
    self.print_clusters(clusters[:5])

    if self.DEBUG:
      self.boxrenderer = JitteredClusterRenderer('/tmp/boxes.pdf')
      self.boxrenderer.plot_clusters(removed_clusters, color="grey")
      for frontier in frontiers:
        self.boxrenderer.plot_clusters(frontier, color="red")
      self.renderer.new_page()
      self.renderer.plot_inf_curves(clusters)
      print self.get_frontier.stats.items()
      print self.get_frontier.heap.stats.items()


    start = time.time()
    iteridx = 1
    seen = set()
    clusters_set = set()
    for frontier in frontiers:
      for c in frontier:
        c.c_range = list(self.c_range)

      versionid = 0
      while len(frontier) > 0:
        self.learner.update_status("expanding frontier, iter %d" % iteridx)
        iteridx += 1

        if self.DEBUG:
          self.renderer.new_page()
          self.renderer.plot_inf_curves(frontier, color='grey', alpha=.2)
          self.renderer.set_title("frontier expansion %d" % iteridx)
          print "frontier"
          self.print_clusters(frontier)

        new_clusters, removed_clusters = self.expand_frontier(frontier, seen, None)

        if self.DEBUG:
          self.renderer.plot_inf_curves(removed_clusters, color="grey")
          self.renderer.plot_inf_curves(new_clusters)
          print "\nnew clusters"
          self.print_clusters(new_clusters)

        if (not new_clusters.difference(frontier)) or (time.time() - start) > 60:
          clusters_set.update(removed_clusters)
          clusters_set.update(new_clusters)
          break

        self.adj_graph.remove(removed_clusters, versionid)
        self.adj_graph.insert(new_clusters, versionid)
        self.adj_graph.sync()
        versionid += 1
        frontier = new_clusters

    print "returning %d clusters total!" % len(clusters_set)
    if self.DEBUG:
      self.renderer.new_page()
      self.renderer.plot_inf_curves(clusters_set)
      self.renderer.set_title("final frontier")
      self.renderer.close()
      self.boxrenderer.close()
      self.print_clusters(clusters_set)

    self.learner.merge_stats(self.get_frontier.stats, 'frontier_')
    self.learner.merge_stats(self.get_frontier.heap.stats, 'inter_heap_')
    return list(clusters_set)
  


  @instrument
  def expand_frontier(self, frontier, seen, version=None):
    """
    Return (newclusters, rmclusters)
    """
    newclusters = set(frontier)
    for cluster in frontier:
      merges = self.expand(cluster, seen, version=None)
      #for c in rms: c.c_range = list(self.c_range)
      #newclusters.update(rms)
      newclusters.update(merges)

    for cluster in newclusters:
      cluster.c_range = list(self.c_range)

    return self.get_frontier(newclusters)

 

  @instrument
  def dim_merge(self, cluster, dim, dec=None, inc=None, skip=None):
    if dec is not None:
      if round(dec, 1) in self.rejected_cont_vals[(dim, 'dec')]:
        return None
    if inc is not None:
      if round(inc, 1) in self.rejected_cont_vals[(dim, 'inc')]:
        return None
    merged = Merger.dim_merge(self, cluster, dim, dec, inc, skip)
    if merged:
      merged.c_range = list(self.c_range)
      merged.inf_func = self.learner.create_inf_func(merged)
    return merged


  @instrument
  def disc_merge(self, cluster, dim, vals, skip=None):
    # reject if union is a superset of anything in 
    # rejected_disc_vals
    vals = set(vals)
    vals.update(cluster.discretes.get(dim, ()))
    for subset in self.rejected_disc_vals[dim]:
      if vals.issuperset(subset):
        return None

    merged = Merger.disc_merge(self, cluster, dim, vals, skip)
    if merged:
      merged.c_range = list(self.c_range)
      merged.inf_func = self.learner.create_inf_func(merged)
    return merged


  @instrument
  def expand(self, c, seen, version=None):
    """
    Returns a frontier of clusters expanded from c that
    are possible optimals in _some_ c range

    XXX: optimization could be a minimum c range a cluster must be
          a candidate over
    """
    _logger.debug("expand\t%s", str(c.rule)[:100])
    start = time.time()
    self.rejected_disc_vals = defaultdict(list)
    self.rejected_cont_vals = defaultdict(set)
    cur_bests = set([c])
    ret = set()
    rms = set()
    all_merges = set()
    self.i = 0
    while cur_bests:
      if not self.DEBUG and (time.time() - start) > 30:
        ret.update(cur_bests)
        break

      cur = cur_bests.pop()
      if cur.bound_hash in seen: 
        _logger.debug("seen \t%s", str(cur.rule)[:100])
        continue
      seen.add(cur.bound_hash)
      _logger.debug("expand\tsub\t%.3f-%.3f\t%s", cur.c_range[0], cur.c_range[1],  str(cur.rule)[:100])


      cur_seen = set()
      dim_to_bests = self.greedy_expansion(cur, cur_seen, version=version)
      seen.update(cur_seen)

      merges = list()
      map(merges.extend, dim_to_bests.values())
      all_merges.update(map(hash, merges))
      merges = set(filter(lambda c: r_vol(c.c_range), merges))
      if cur in merges:
        _logger.debug("cur added to bests")
        ret.add(cur)
        merges.remove(cur)

      if len(merges) is 0:
        _logger.debug("# better expanded = 0")
        ret.add(cur)
        continue

      combined = set(cur_bests)
      combined.update(merges)
      frontier, losers = self.get_frontier(combined)


      if self.DEBUG:
        self.boxrenderer.new_page()
        self.boxrenderer.plot_clusters([l for l in losers if l not in merges], color="grey", alpha=0.2)
        self.boxrenderer.plot_clusters([m for m in merges if m in losers], color='blue', alpha=0.2)
        self.boxrenderer.plot_clusters([m for m in merges if m in frontier], color='red', alpha=0.3)
        self.boxrenderer.plot_clusters([cur], color="black", alpha=0.7)
        self.boxrenderer.set_title("expand iter %d\n%s" % (self.i, str(c.rule)))
        self.i += 1



      if cur_bests.issuperset(frontier):
        _logger.debug("frontier subset of curbests")
        ret.add(cur)
        continue
      
      for merged in frontier.difference(cur_bests):
        rms.update(merged.parents)
      self.adj_graph.insert(cur_bests)
      seen.update([c.bound_hash for c in rms])

      cur_bests = frontier
      cur_bests.difference_update([cur])
      _logger.debug("\t%d in frontier", len(cur_bests))




    return ret
    ret, more_rms = self.get_frontier(ret)
    rms.update(more_rms)
    return ret, rms

class RangeMerger(BaseRangeMerger):

  @instrument
  def greedy_expansion(self, cur, seen, version=None):
    expansions = self.expand_candidates(cur, seen, version=version)

    dim_to_bests = defaultdict(set)
    for dim, direction, g in expansions:
      dim_bests = self.expand_dim(cur, dim, direction, g, seen)
      #if cur in dim_bests: dim_bests.remove(cur)
      dim_to_bests[(dim, direction)] = dim_bests

    # cross product between inc and dec of each dimension

    return dim_to_bests
  
  @instrument
  def expand_dim(self, cur, dim, direction, g, seen):
    """
    Args
      dim: if direction == 'disc', dim is the attr name
           else dim is the index into cluster.cols
    """
    if True or direction == 'disc':
      cands = []
      bests = set([cur])
      prev_card = None
      cands = list(g)
      bests.update(cands)
      bests, _ = self.get_frontier(bests)
      for cand in cands:
        if cand not in bests:
          seen.add(cand.bound_hash)
    else:
      cands = []
      bests = set([cur])
      for cand in g:
        cands.append(cand)
        bests, _ = self.get_frontier(bests.union(set([cand])))
        if cand not in bests:
          seen.add(cand.bound_hash)
          break
      bests = [cur] + cands
    bests = list(bests)

    # update rejection state
    for c in cands:
      if c in bests: continue
      if direction == 'disc':
        self.rejected_disc_vals[dim].append(set(c.discretes[dim]))
      if direction == 'inc':
        v = c.bbox[1][dim]
        self.rejected_cont_vals[(dim, direction)].add(v)
      if direction == 'dec':
        v = c.bbox[0][dim]
        self.rejected_cont_vals[(dim, direction)].add(v)

    if self.DEBUG:
      if direction == 'disc':
        name = dim[:10]
        s = ','.join([str(len(c.discretes[dim])) for c in bests])
        ss = ','.join([str(c.discretes[dim]) for c in cands])
        if 'subregion' in str(cur):
          for c in cands:
            isbest = False
            if c in bests:
              isbest = True
            _logger.debug('\tbest? %s\tcand:\t%s', isbest, c)
      else:
        name = cur.cols[dim][:10]
        if bests:
          if direction == 'inc':
            s = ','.join(["%.4f" % c.bbox[1][dim] for c in bests])
            s = '%.4f - %s' % (bests[0].bbox[0][dim], s)
          else:
            s = ','.join(["%.4f" % c.bbox[0][dim] for c in bests])
            s = '%s - %.4f' % (s, bests[0].bbox[1][dim])
        else:
          s = '---'
        ss = ''
      _logger.debug("\t%s\t%s\t%d bests\t%d candidates", name, direction, len(bests), len(cands))
      _logger.debug("\tbests\t\t\t\t%s", s)
      _logger.debug("\tcands\t\t\t\t%s", ss)


    return bests


class RangeMerger2(BaseRangeMerger):

  def pick_expansion_vals(self, cluster, dim, direction, vals):
    if len(vals) == 0: return vals
    if direction == 'disc': 
      return vals
      return np.random.choice(vals, min(len(vals), 6), replace=False)

    vals = random.sample(vals, min(4, len(vals)))
    vals.sort(reverse=(direction == 'dec'))

    if direction == 'inc':
      baseval = cluster.bbox[1][dim]
    else:
      baseval = cluster.bbox[0][dim]

    rng = self.learner.cont_dists[cluster.cols[dim]]
    min_step = r_vol([rng.min, rng.max]) * 0.005
    vals = filter(lambda v: abs(v-baseval) > min_step, vals)

    if len(vals) <= 4:
      return vals
    vals = random.sample(vals, 4)
    return vals


  @instrument
  def greedy_expansion(self, cluster, seen, version=None):
    curset = set([cluster])
    for dim, direction, vals in self.dims_to_expand(cluster, seen, version=version):
      if self.DEBUG:
        self.boxrenderer.new_page()
        self.boxrenderer.plot_clusters(curset, color='blue', alpha=0.3)
        _logger.debug("%s\t%s\t%s", dim, direction, str(vals))


      tmp = set()
      realvals = self.pick_expansion_vals(cluster, dim, direction, vals)
      if direction != 'disc':
        _logger.debug("# to expand\t%s\t%s\t%d -> %d\t%s -> %s", 
            cluster.cols[dim][:10], direction, len(vals), 
            len(realvals), str(vals), str(realvals))
      if direction == 'inc':
        for inc in realvals:
          tmp.update([self.dim_merge(c, dim, None, inc, seen) for c in curset])
      elif direction == 'dec':
        for dec in realvals:
          tmp.update([self.dim_merge(c, dim, dec, None, seen) for c in curset])
      else:
        for disc in realvals:
          tmp.update([self.disc_merge(c, dim, disc) for c in curset])

      tmp = filter(bool, tmp)
      _logger.debug("# actual expansions\t%d using %d of %d vals\t%d in curset", len(tmp), len(realvals), len(vals), len(curset))
      if not tmp:
        continue
      for c in curset:
        c.c_range = list(self.c_range)
      tmp.extend(curset)
      curset,_ = self.get_frontier(tmp)
      seen.update([c.bound_hash for c in _])


      # update rejection state
      for v in vals:
        if direction == 'disc':
          self.rejected_disc_vals[dim].append(set(v))#c.discretes[dim]))
        if direction == 'inc':
          #v = c.bbox[1][dim]
          self.rejected_cont_vals[(dim, direction)].add(round(v, 1))
        if direction == 'dec':
          #v = c.bbox[0][dim]
          self.rejected_cont_vals[(dim, direction)].add(round(v, 1))



      if self.DEBUG:
        self.boxrenderer.plot_clusters(curset, color='red', alpha=0.3)
        self.boxrenderer.plot_clusters(_, color='grey', alpha=0.2)
        self.boxrenderer.set_title("greedy expansion\n%s" % cluster)


    ret = { 'dim': curset }
    return ret






import os
from sqlalchemy import create_engine
from common import *

def print_clusters(sub, dim, clusters=[], tuples=[], title=''):
  xattr = 'a_%d' % dim
  yattr = 'a_%d' % (dim+1)

  for cluster in clusters:
    bbox = tuple(map(list, zip(*cluster.bbox)))
    # we want bounds for attrs a_dim, a_dim+1, but bbox cols
    # may not be ordered as we like
    if xattr in cluster.cols:
      xidx = cluster.cols.index(xattr)
      x = bbox[xidx]
      x[0] = max(0, x[0])
      x[1] = min(100, x[1])
    else:
      x = [0, 100]

    if yattr in cluster.cols:
      yidx = cluster.cols.index(yattr)
      y = bbox[yidx]
      y[0] = max(0, y[0])
      y[1] = min(100, y[1])
    else:
      y = [0, 100]

    c = cm.jet(cluster.error)
    r = Rect((x[0], y[0]), x[1]-x[0], y[1]-y[0], alpha=min(1., max(0.2,cluster.error)), ec=c, fill=False, lw=1.5)
    sub.add_patch(r)

  if tuples:
    cols = zip(*tuples)
    xs, ys, cs = cols[dim], cols[dim+1], cols[-2]
    cs = np.array(cs) / 100.
    sub.scatter(xs, ys, c=cs, alpha=0.5, lw=0)

  sub.set_ylim(-5, 105)
  sub.set_xlim(-5, 105)
  sub.set_title(title)



def print_all_clusters(pp, db, tablename, learner, c):
  try:
    all_clusters = [cluster.clone() for cluster in learner.all_clusters]
    all_clusters = normalize_cluster_errors(all_clusters)
    clusters = [cluster.clone() for cluster in learner.final_clusters]
    clusters = normalize_cluster_errors(clusters)
    best_clusters = sorted(clusters, key=lambda c: c.error, reverse=True)
    best_clusters = best_clusters[:2]
    best_clusters[0].error = 1
    tuples = get_tuples_in_bounds(db, tablename, [], 'g = 7')

    for cl in clusters:
      print str(cl), cl.c_range

    for dim in xrange(len(tuples[0])-4):
      suffix = "%.4f dim %d" % (c, dim)
      fig = plt.figure(figsize=(12, 4))
      print_clusters(fig.add_subplot(1, 3, 1), dim, all_clusters, tuples=tuples,title="merged %s" % suffix)
      print_clusters(fig.add_subplot(1, 3, 2), dim, clusters, tuples=tuples,title="merged %s" % suffix)
      print_clusters(fig.add_subplot(1, 3, 3), dim, best_clusters, tuples=tuples, title="best %s" % suffix)
      plt.savefig(pp, format='pdf')

  except Exception as e:
    import traceback
    traceback.print_exc()
    pdb.set_trace()
    pass



def run(pp, cutoff, **params):
  dataset = params['dataset']
  test_datas = get_test_data(datasetnames[dataset])
  tablename = test_datas[-1]
  dbname = test_datas[0]
  db = create_engine('postgresql://localhost/%s' % dbname)

  costs, rules, all_ids, table_size, learner = run_experiment(dataset, **params)
  cost = costs['cost_total']
  ft = learner.full_table
  print len(ft)
  truth = [int(row['id'].value) for row in ft if row['v'] >= cutoff]
  all_stats = [compute_stats(ids, truth,  table_size) for ids in all_ids]

  stats, rule, ids = tuple(zip(all_stats, rules, all_ids)[0])
  data = tuple([tablename,params['c'],cost]+list(stats))
  print "stats:%s,c(%.3f),cost(%.2f),%.6f,%.6f,%.6f,%.6f" % data
  print 'stats:%s'% str(sdrule_to_clauses(rule.simplify())[0])
  print_all_clusters(pp, db, tablename, learner, params['c'])
  return costs


def warmup(dim, cutoff, **kwargs):
  dataset = "data_%d_%d_1000_0d50_%duo" % (dim, dim, cutoff)
  params = {
    'klass':BDT, 
    'nbadresults' : 10,
    'epsilon':0.005,
    'tau':[0.1, 0.5],
    'p' : 0.7,
    'l':.5,
    'min_pts' : 10,
    'min_improvement':.01,
    'granularity':15,
    'max_wait':1,
    'naive':False,
    'use_mtuples':False,
    'use_cache': False
  }
  params.update(kwargs)

  ft, bts, gts, truth, aggerr, cols = get_parameters(dataset, **params)
  params.update({
    'aggerr' : aggerr,
    'cols' : cols,
    'tablename' : dataset,
    'dataset' : dataset
  })
  learner = BDT(**params)
  learner.setup_tables(ft, bts, gts, **params)
  learner.get_partitions(ft, bts, gts, **params)


def run_cache(dim, cutoff, cs, **kwargs):
  dataset = kwargs.get('dataset', "data_%d_%d_1000_0d50_%duo" % (dim, dim, cutoff))

  params = {
    'klass':BDT, 
    'nbadresults' : 10,
    'epsilon':0.005,
    'tau':[0.1, 0.5],
    'p' : 0.7,
    'l':.5,
    'min_pts' : 10,
    'min_improvement':.01,
    'granularity':15,
    'max_wait':20,
    'naive':False,
    'use_mtuples':False,
    'use_cache': False,
    dataset: dataset
  }
  params.update(kwargs)


  pp = PdfPages('figs/topdown_all_%s.pdf' % str(dataset))
  cost_dicts = []
  for c in cs:
    params['c'] = c
    cost_dict = run(pp, cutoff, **params)
    cost_dicts.append(cost_dict)

  pp.close()
  return cost_dicts


def reset_cache():
  try:
    os.system('rm dbwipes*.cache')
  except:
    pass


if __name__ == '__main__':
  np.seterr(all='raise')
  if len(sys.argv) < 4:
    print "python run_cache_experiments.py [dimensions] [30|80] [cache? 0|1] [list of cs values]"
    print "cs values defaults to [.5, .4, .3, .2, .1, .05, 0]"
    sys.exit()

  dim = int(sys.argv[1])
  uo = int(sys.argv[2])
  cache = bool(int(sys.argv[3]))
  cs = map(float, sys.argv[4:])
  if not cs:
    cs = [.5, .4, .3, .2, .1, 0.05, 0.0] 

  #reset_cache()
  #cachecost_dicts = run_cache(dim, uo, cs, l=0.95, tree_alg='rt', klass=NDT, use_cache=cache, tau= [0.1, 0.5])

  cachecost_dicts = run_cache(dim, uo, cs, l=0.85, tree_alg='rt', klass=BDT, 
      epsilon=0.001, use_cache=cache, tau= [0.02, 0.5],
      c_range=[0.01, 0.7],
      dataset='data2clust_2_2_2k_vol20_uo80')




  print "c,total,partbad,partgood,split,merge,cache"
  for c, cd in zip(cs, cachecost_dicts):
    print "%.4f,%.4f,%.4f,%.4f,%.4f,%.4f,%d" % (
      c, 
      cd.get('cost_total', -1),
      cd.get('cost_partition_bad', -1),
      cd.get('cost_partition_good', -1),
      cd.get('cost_split', -1),
      cd.get('cost_merge', -1), 
      cache
    )



#
# numerous helper methods for comparing influence dominance
# between predicates across a range of c (lambda in job talk)
# values
#
# 

import json
import math
import pdb
import random
import numpy as np
import sys
import time
from collections import *
from scipy.optimize import fsolve
sys.path.extend(['.', '..'])

from scorpion.util import *

INF = float('inf')


class Frontier(object):
  _id = 0

  """
  Iteratively look for the skyline of a list of influence functions
  by identifying intersection points while scanning from left to right

  The module is actually passed lists of Cluster objects, which implement
  an influence function: 

      c.inf_func(c_value) -> influence value

  while heap:
    if init
      for the top cluster at c_range[0]
        find intersection points with all other clusters
        pick next intersection point, add to heap
    else
      swap at intersection point
      find intersection points with all other clusters
      pick next intersection point, add to heap

  NOTE: picks _at most one_ function on the frontier, not _all_
        functions on the frontier
  """
  def __init__(self, c_range, min_granularity=0):
    """
    Args
      min_granularity: minimum distance between adjacent intersections
    """
    self.c_range = c_range
    self.min_granularity = min_granularity
    self.heap = Intersection(c_range)
    self.stats = defaultdict(lambda: [0,0])
    self.id = Frontier._id 
    Frontier._id += 1

  def __call__(self, clusters):
    """
    Return:
      (frontier, removed)

    """
    if not clusters or len(clusters) <= 1: 
      return set(clusters), set()

    frontier = self.get_frontier(clusters)
    ret = self.frontier_to_clusters(frontier)
    rms = set(clusters)
    if frontier:
      rms.difference_update(zip(*frontier)[0])
    for rm in rms:
      rm.c_range = [rm.c_range[0], rm.c_range[0]]
    return ret, rms

  
  @instrument
  def frontier_to_clusters(self, frontier):
    """
    Return a list of clusters with proper bounds
    avoid cloning cluster if possible
    Args:
      list of (cluster, start, end)
    """
    ret = set()
    for cur, start, end in frontier:
      c = cur.clone(copy_rule=True)
      c.c_range = r_intersect([start, end], cur.c_range)
      ret.add(c)

    return ret

  
  def intersect_ranges(self, c1, c2):
    """
    @deprecated
    Given two clusters, split them up by the bounds where they are each 
    optimal
    """
    frontier = self.get_frontier([c1, c2])
    d = defaultdict(list)
    for c, s, e in frontier:
      d[c.id].append((s, e))
    
    c1s, c2s = [], []

    for (s, e) in d[c1.id]:
      c = c1.clone()
      c.c_range = r_intersect([s, e], c1.c_range)
      c1s.append(c)
    if not c1s:
      c1.c_range = [c1.c_range[0], c1.c_range[0]]

    
    for (s, e) in d[c2.id]:
      c = c2.clone()
      c.c_range = r_intersect([s, e], c2.c_range)
      c2s.append(c)
    if not c2s:
      c2.c_range = [c2.c_range[0], c2.c_range[0]]

    return c1s, c2s



  @instrument
  def get_frontier(self, clusters):
    """
    Groups clusters by non-overlapping c_ranges first
    """
    groups = self.group_by_crange(clusters)
    ret = []
    for group in groups:
      ret.extend(self._get_frontier(group))
    return ret
 
  def _get_frontier(self, clusters):
    """
    Core method that computes increasing intersections
    Rteurn:
      list of (cluster, start, end) 
    """
    # XXX: doesn't work for intersections of >2 models
    if not clusters: return []


    start = time.time()
    init_intersection = self.c_range[0]
    topcluster = max(clusters, key=lambda c: c.inf_func(init_intersection))
    heap = self.heap
    start = time.time()

    ignore = set()
    frontier = [(init_intersection, topcluster)]
    while frontier:
      cur_inter, c1 = frontier[-1]

      if cur_inter + self.min_granularity > self.c_range[1]:
        break

      start = time.time()
      # collect all intersections
      roots = []
      for c2 in clusters:
        if c1 == c2: continue
        #if c2 in ignore: continue

        root = heap(c1, c2, cur_inter + self.min_granularity)
        if root is not None:
          roots.append(root)
        else:
          # how could c2 be best if it doesn't intersect with the top?
          ignore.add(c2)

      cost = time.time() - start
      self.stats['root_scan'][0] += cost
      self.stats['root_scan'][1] += 1

      if not roots: break

      # pick cluster that will be top after the next intersection
      nextroot = min(roots)
      nextc = max(clusters, key=lambda c: c.inf_func(nextroot+1e-10))
      if nextc == c1:
        break

      frontier.append((nextroot, nextc))

    # extend last cluster to the end of the c_range
    frontier.append((self.c_range[1], frontier[-1][1]))

    ret = []
    for idx, (start, c) in enumerate(frontier[:-1]):
      ret.append((c, start, frontier[idx+1][0]))
    return ret


  def group_by_crange(self, clusters):
    """
    return disjoint groups of clusters wrt their c_ranges
    """
    clusters = sorted(clusters, key=lambda c: c.c_range[0])

    # list of: (group, union_c_range)
    groups, ranges = [], []
    for c in clusters:
      found = False
      for idx, (group, c_range) in enumerate(zip(groups, ranges)):
        if r_vol(r_intersect(c_range, c.c_range)):
          group.append(c)
          ranges[idx] = r_union(c_range, c.c_range)
          found = True
          break
      if not found:
        groups.append([c])
        ranges.append(list(c.c_range))
    return groups



class ContinuousFrontier(Frontier):
  """
  Maintains the frontier across a "stream" of cluster objects

  The update() method updates the internal set of clusters on the 
  frontier.
  """


  def __init__(self, c_range, min_granularity=0):
    """
    Args
      min_granularity: minimum distance between adjacent intersections
    """
    Frontier.__init__(self, c_range, min_granularity=min_granularity)

    # persistent frontier across all clusters this has seen
    self.frontier = set()
    self.seen = set()

  def __contains__(self, c):
    if c is None: return False
    if hasattr(c, 'bound_hash'):
      return c.bound_hash in set([c2.bound_hash for c2 in self.frontier])
      #self.seen
    return False

  @instrument
  def intersections_with_frontier(self, cluster):
    for c in self.frontier:
      if r_vol(r_intersect(c.c_range, cluster.c_range)):
        yield c


  @instrument
  def update(self, clusters):
    adds, rms, new_spans = self.probe(clusters)

    new_clusters = self.frontier_to_clusters(new_spans)
    self.frontier.difference_update(rms)
    self.frontier.update(new_clusters)
    self.seen.update([c.bound_hash for c in clusters])

    adds_hashes = set([c.bound_hash for c in clusters])
    adds = [c for c in self.frontier if c.bound_hash in adds_hashes]
    adds = set(adds)
    return adds, rms


  @instrument
  def probe(self, clusters):
    """
    merge clusters in argument with existing frontier
    """
    new_spans = []
    beaten_counts = defaultdict(lambda: 0)
    arg_frontier = set(self(clusters)[0])

    for c in arg_frontier:
      if c.bound_hash in self.seen:
        continue

      c_new_spans = []
      for cand_idx, cand in enumerate(self.frontier):
        if r_vol(r_intersect(c.c_range, cand.c_range)) == 0:
          continue
        
        valid_bound = r_intersect(c.c_range, cand.c_range)
        is_c_best = c.inf_func(valid_bound[0]) > cand.inf_func(valid_bound[0])
        # the following is indexed into using is_c_best
        pair = [cand, c]

        cur_inter = valid_bound[0]
        while cur_inter < valid_bound[1]:
          root = self.heap(c, cand, cur_inter)
          if root is None:
            root = valid_bound[1]

          root = min(valid_bound[1], root)
          if root <= cur_inter: break

          better_cluster = pair[is_c_best]
          c_new_spans.append((better_cluster, cur_inter, root))
          beaten_counts[pair[not is_c_best]] += 1

          cur_inter = root
          is_c_best = not is_c_best

      if not c_new_spans:
        new_spans.append((c, c.c_range[0], c.c_range[1]))
      else:
        new_spans.extend(c_new_spans)

    rms = set([c for c, n in beaten_counts.iteritems() if n > 0])
    new_spans = filter(lambda s: s[2] > s[1], new_spans)
    new_spans = filter(lambda s: s[0] in arg_frontier or s[0] in rms, new_spans)
    adds = set([c for c, minc, maxc in new_spans if c in arg_frontier])

    return adds, rms, new_spans




class CheapFrontier(Frontier):
  """
  Sampling based approximate frontier.  Computes influence values
  at a fixed number of c values for each cluster and uses those to
  pick the frontier.

  In practice, >40 c values gets pretty good results.
  Ends up being faster than the above root-finding-based approaches
  """
  buckets_cache = {}

  def __init__(self, c_range, min_granularity=0, K=3, nblocks=100):
    """
    Args
      min_granularity: minimum distance between adjacent intersections
      K: select top-k at each bucket
    """
    Frontier.__init__(self, c_range, min_granularity=min_granularity)


    self.seen = set()
    self.seen_clusters = set()

    self.nblocks = nblocks
    self.buckets = CheapFrontier.compute_normalized_buckets(nblocks, self.seen_clusters)
    self.buckets = self.buckets * r_vol(c_range) + c_range[0]
    self.nblocks = len(self.buckets)
    self.bests = defaultdict(list)   # bucket -> clusters
    self.K = K

    self.frontier = []
    self.frontier_hashes = []
    self.frontier_infs = []
    self.threshold = np.zeros(self.nblocks).astype(float)

  @staticmethod
  def compute_normalized_buckets(nblocks, clusters=[]):
    """
    return distribution of x buckets, normalized to be between 0 and 1
    """
    if nblocks in CheapFrontier.buckets_cache:
      return CheapFrontier.buckets_cache[nblocks]

    def mkf(pairs):
      return np.vectorize(lambda c: np.mean([t/pow(b,c) for t,b in pairs]) )

    inf_funcs = []
    for c in clusters:
      inf_funcs.append(c.inf_func)

    while len(inf_funcs) < 50:
      tops = np.random.rand(6) * 20
      bots = np.random.rand(6) * 20 + 1.
      pairs = zip(tops, bots)
      inf_funcs.append(mkf(pairs))

    nblocks = int(nblocks)
    xs = np.arange(nblocks * 2).astype(float) / (nblocks * 2.)
    all_infs = np.array([inf_func(xs) for inf_func in inf_funcs])
    medians = np.percentile(all_infs, 50, axis=0)
    block = (medians.max() - medians.min()) / nblocks
    optxs = []
    ys = []
    prev = None
    for idx, v in enumerate(medians):
      if prev == None or abs(v - prev) >= block:
        optxs.append(xs[idx])
        ys.append(v)
        prev = v
    optxs = np.array(optxs)
    optxs -= optxs.min()
    optxs /= optxs.max()

    if len(clusters) > 30:
      CheapFrontier.buckets_cache[nblocks] = optxs
    return optxs

  @instrument
  def cluster_infs(self, c):
    infs = c.inf_func(self.buckets)
    infs[np.abs(infs) == INF] = -1
    return infs

  @instrument
  def compute_thresholds(self, all_infs, thresholds=None, K=None):
    if thresholds is None:
      thresholds = np.zeros(self.nblocks).astype(float)
      thresholds[:] = -1e100
    if K is None:
      K = self.K

    ret = []
    for bidx, bucket in enumerate(self.buckets):
      bucket_infs = all_infs[:, bidx]
      if len(bucket_infs) <= K:
        thresh = bucket_infs.min()
      else:
        idx = np.argpartition(bucket_infs, -K)[-K]
        thresh = bucket_infs[idx]
      thresholds[bidx] = max(thresholds[bidx], thresh)

    return thresholds


  @instrument
  def _get_frontier(self, clusters, K=None):
    if len(clusters) == 0:
      return []
    if len(clusters) == 1:
      return [ (clusters[0], self.c_range[0], self.c_range[1]) ]

    clusters = list(clusters)
    all_infs = [self.cluster_infs(c) for c in clusters]
    try:
      all_infs = np.array(all_infs)
    except Exception as e:
      print e
      print all_infs
    thresholds = self.compute_thresholds(all_infs, K=K)


    # all_infs is a clusters x buckets matrix
    # with a 1 in a cell if a cluster is "best" for
    # that bucket
    return self.all_infs_to_spans(clusters, all_infs, thresholds)

  def all_infs_to_spans(self, clusters, all_infs, thresholds=None):
    ret = []
    for idx, c in enumerate(clusters):
      infs = all_infs[idx]
      ret.extend(self.infs_to_spans(c, infs, thresholds))
    return ret

  @instrument
  def infs_to_spans(self, cluster, infs, thresholds=None):
    if thresholds is None: 
      thresholds = self.thresholds

    ret = []
    passes = infs >= thresholds
    sidx = None
    for bidx in xrange(len(self.buckets)):
      if passes[bidx] == 1:
        if sidx is None:
          sidx = bidx
      else:
        if sidx is not None:
          start = self.buckets[sidx]
          end = self.buckets[bidx-1]
          if bidx < len(self.buckets):
            end = self.buckets[bidx]
          else:
            end = self.c_range[1]
          #print '%.2f-%.2f\t%s' % (start, end, cluster.rule)
          ret.append((cluster, start, end))
          sidx = None

    if sidx is not None:
      start = self.buckets[sidx]
      end = self.c_range[1]
      #print '%.2f-%.2f\t%s' % (start, end, cluster.rule)
      ret.append((cluster, self.buckets[sidx], self.c_range[1]))
    return ret

  def improvement(self, cluster):
    """
    cluster.influence - frontier for each bucket
    """
    infs = self.cluster_infs(cluster)
    if len(self.frontier_infs) == 0:
      return infs
    return infs - self.frontier_infs.max(axis=0)

  @instrument
  def update(self, clusters, K=None):
    clusters = self(clusters)[0]
    if len(clusters) == 0: 
      return set(), set()
    
    clusters = [c for c in clusters if c.bound_hash not in self.seen]

    arg_infs = [self.cluster_infs(c) for c in clusters]
    arg_infs.extend(self.frontier_infs)
    all_infs = np.array(arg_infs)
    all_clusters = list(clusters)
    all_clusters.extend(self.frontier)
    self.thresholds = self.compute_thresholds(all_infs, K=K)
    spans = self.all_infs_to_spans(all_clusters, all_infs, self.thresholds)

    span_hashes = set([tup[0].bound_hash for tup in spans])
    in_hash = lambda c: c.bound_hash in span_hashes

    rms = [c for c in self.frontier if not in_hash(c)]
    adds = [c for c in clusters if in_hash(c)]
    hash2infs = {c.bound_hash : infs for c, infs in zip(all_clusters, all_infs)}

    self.frontier = self.frontier_to_clusters(spans)
    self.frontier_hashes = [c.bound_hash for c in self.frontier]
    self.frontier_infs = []
    for h, c in izip(self.frontier_hashes, self.frontier):
      self.frontier_infs.append(hash2infs.get(h, self.cluster_infs(c)))
    self.frontier_infs = np.array(self.frontier_infs)
    self.frontier_hashes = set(self.frontier_hashes)

    for c in clusters:
      self.seen.add(c.bound_hash)
      if len(self.seen_clusters) < 50:
        self.seen_clusters.add(c)
    return adds, rms

  def __contains__(self, c):
    if c is None: return False
    if hasattr(c, 'bound_hash'):
      return c.bound_hash in self.frontier_hashes
    return False


class Intersection(object):
  """
  Computes and Caches intersection points between cluster curves
  """
  def __init__(self, bound):
    self.cache = defaultdict(list)
    self.bound = bound
    self.tokey = lambda c1, c2: (min(c1.bound_hash, c2.bound_hash), max(c1.bound_hash, c2.bound_hash))
    self.zero_thresh = 1e-7
    self.stats = defaultdict(lambda: [0, 0])
    self.rng = np.arange(5) / 5.

  @instrument
  def __call__(self, c1, c2, minv=None, maxv=None):
    """
    Return next intersection of the two models
           that is greater than minv
    """
    if c1 is None: return None
    if c2 is None: return None
    if abs(c1.error) == float('inf'): return None
    if abs(c2.error) == float('inf'): return None


    bound = list(self.bound)
    if minv is not None:
      bound[0] = minv
    if maxv is not None:
      bound[1] = maxv


    if bound[0] == bound[1]:
      return None

    # XXX: make a REALLY big assumption that inf functions
    #      are convex/concave, thus checking the values at a small
    #      number of points is enough to assess if they intersect
    #      at all
    xs = (self.rng * r_vol(bound)) + bound[0]
    start = time.time()
    c1s = c1.inf_func(xs)
    c2s = c2.inf_func(xs)
    try:
      nbools = (c1s < c2s).sum()
      nequals = (c1s == c2s).sum()
    except:
      return None
    cost = time.time() - start
    self.stats['aprox_check'][0] += cost
    self.stats['aprox_check'][1] += 1
    if nbools == len(xs) or nbools == 0:
      return None

    # the curves are identical
    if nequals > 2:
      return None # wrong to return none

    f = np.vectorize(lambda v: np.abs(c1.inf_func(v) - c2.inf_func(v)))

    key = self.tokey(c1, c2)
    if key in self.cache:
      roots = self.cache[key]
    else:
      start = time.time()
      roots = fsolve(f, bound[0], maxfev=50)
      roots.sort()
      cost = time.time() - start
      self.stats['fsolve_calls'][0] += cost
      self.stats['fsolve_calls'][1] += 1


    if key not in self.cache:
      self.cache[key] = roots

    roots = filter(lambda v: v > bound[0] and v <= bound[1], roots)
    roots = filter(lambda v: abs(f(v)) < self.zero_thresh, roots)
    if roots:
      return roots[0]
    return None






import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

from itertools import chain


from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *

from basic import Basic
from merger import Merger
from grouper import Grouper, Blah

_logger = get_logger()


class NaiveGrouper(Grouper):

    def __call__(self, attrs, valid_groups):
        """ every group is valid!!!"""

        start = time.time()
        bad_table_rows = []
        good_table_rows = []
        for table in self.mr.bad_tables:
            bad_table_rows.append(self.table_influence(attrs, table))
        for table in self.mr.good_tables:
            good_table_rows.append(self.table_influence(attrs, table))
        #print "scan time\t", (time.time() - start)

        def get_infs(all_table_rows, err_funcs, g, bmaxinf):
            ret = []
            counts = []
            maxinf = None
            for idx, ef, table_rows in zip(range(len(err_funcs)), err_funcs, all_table_rows):
                rows = table_rows.get(g, [])
                if rows:
                    if bmaxinf:
                        cur_maxinf = self.influence_tuple(max(rows, key=lambda row: self.influence_tuple(row, ef)), ef)
                        if not maxinf or cur_maxinf > maxinf:
                            maxinf = cur_maxinf

                    ret.append(ef(rows))
                    counts.append(len(rows))
            return ret, counts, maxinf


        start = time.time()
        for g in self.keys_iter(attrs):
            bds, bcs, maxinf = get_infs(bad_table_rows, self.mr.bad_err_funcs, g, True)
            gds, gcs, _ = get_infs(good_table_rows, self.mr.good_err_funcs, g, False)
            if not bcs:
                continue
            yield Blah(attrs, g, bds, bcs, gds, gcs, maxinf, self.mr, self)
        #print "comp infs\t", (time.time() - start)


    def influence_tuple(self, row, ef):
        if row[self.mr.SCORE_ID].value == -1e10000000000:
            influence = ef((row,))
            row[self.mr.SCORE_ID] = influence
        return row[self.mr.SCORE_ID].value


    def keys_iter(self, attrs):
        seen = set()
        for row in self.data:
            group = tuple([self.gbfuncs[attr](row[attr]) for attr in attrs])
            if group in seen:
                continue
            seen.add(group)
            yield group 


    def table_influence(self, attrs, table):
        """scan table once and group tuples by their respective groups"""
        groups = defaultdict(list)
        for row in table:
            group = tuple([self.gbfuncs[attr](row[attr]) for attr in attrs])
            groups[group].append(row)
        return groups


    def merge_groups(self, prev_groups):
        """
        prev_groups: attributes -> groups
        attributes are sorted
        group: attr -> idx
        """
        start = time.time()
        attrs_list = prev_groups.keys()
        for a_idx, attrs1 in enumerate(attrs_list):
            sattrs1 = set(attrs1)
            for attrs2 in attrs_list[a_idx+1:]:
                pgroup1, pgroup2 = prev_groups[attrs1], prev_groups[attrs2]

                sattrs2 = set(attrs2)
                merged_attrs = tuple(sorted(sattrs1.union(sattrs2)))
                if len(merged_attrs) != len(sattrs1)+1:
                    continue
                yield (merged_attrs, [])
            

